import getpass
import traceback
from collections import OrderedDict
from functools import partial
from importlib import import_module
import os
import multiprocessing

import subprocess
from time import time

import math

from artemis.general.display import equalize_string_lengths
from six import string_types
from six.moves import reduce, xrange
from artemis.experiments.experiment_record import (load_experiment_record, ExpInfoFields,
    ExpStatusOptions, ARTEMIS_LOGGER, record_id_to_experiment_id, get_all_record_ids, get_experiment_dir)
from artemis.experiments.experiments import load_experiment, get_global_experiment_library
from artemis.fileman.config_files import get_home_dir,set_non_persistent_config_value
from artemis.general.hashing import compute_fixed_hash
from artemis.remote.child_processes import SlurmPythonProcess
from artemis.remote.nanny import Nanny
from artemis.general.should_be_builtins import izip_equal, detect_duplicates, remove_common_prefix, memoize, \
    divide_into_subsets


def pull_experiments(user, ip, experiment_names, include_variants=True, need_pass = False):
    """
    Pull experiments from another computer matching the given experiment name.

    :param user:
    :param ip:
    :param experiment_name:
    :param include_variants:
    :return:
    """
    import pexpect
    import sys

    if isinstance(experiment_names, string_types):
        experiment_names = [experiment_names]

    home = get_home_dir()

    # This one works if you have keys set up
    command = ['rsync', '-a', '-m', '-i']\
        +['{user}@{ip}:~/.artemis/experiments/'.format(user=user, ip=ip)] \
        +['{home}/.artemis/experiments/'.format(home=home)]\
        +["--include='**/*-{exp_name}{variants}/*'".format(exp_name=exp_name, variants = '*' if include_variants else '') for exp_name in experiment_names] \
        +["--include='*/'", "--exclude='*'"]

    if not need_pass:
        output = subprocess.check_output(command)
        return output
    else:
        # This one works if you need a password
        password = getpass.getpass("Enter password for {}@{}:".format(user, ip))
        child = pexpect.spawn(' '.join(command))
        code = child.expect([pexpect.TIMEOUT, 'password:'])
        if code == 0:
            print(("Got unexpected output: %s %s" % (child.before, child.after)))
            sys.exit()
        else:
            child.sendline(password)
        output = child.read()
        return output


def load_lastest_experiment_results(experiments, error_if_no_result = True):
    """
    Given a list of experiments (or experiment ids), return an OrderedDict<record_id: result>
    :param experiments: A list of Experiment objects (or strings identifying experiment ID is ok too)
    :param error_if_no_result: Raise an error if an experiment has no completed results.
    :return: OrderedDict<record_id: result>
    """
    experiments = [load_experiment(ex) if isinstance(ex, string_types) else ex for ex in experiments]
    records = [ex.get_latest_record(err_if_none=error_if_no_result, only_completed=True) for ex in experiments]
    record_results = load_record_results([r for r in records if r is not None], err_if_no_result=error_if_no_result)
    experiment_latest_results = OrderedDict((rec.get_experiment_id(), val) for rec, val in record_results.items())
    return experiment_latest_results


def load_record_results(records, err_if_no_result =True, index_by_id = False):
    """
    Given a list of experiment records, return an OrderedDict<record: result>
    :param records: A list of ExperimentRecord objects
    :param err_if_no_result: True to raise an error if a record has no result.  If False, you will just be warned that
        the result does not exist for this record.
    :return:  OrderedDict<ExperimentRecord: result>
    """
    results = OrderedDict()
    for record in records:
        index = record.get_id() if index_by_id else record
        if not record.has_result():
            if err_if_no_result:
                raise Exception('Record {} had no result.'.format(record.get_id()))
            else:
                ARTEMIS_LOGGER.warn('Experiment Record {} had no saved result.  Not including this in results'.format(record.get_id()))
        else:
            results[index] = record.get_result()
    return results


def select_experiments(user_range, exp_record_dict, return_dict=False):
    exp_filter = _filter_experiments(user_range, exp_record_dict)
    if return_dict:
        return OrderedDict((name, exp_record_dict[name]) for name in exp_record_dict if exp_filter[name])
    else:
        return [name for name in exp_record_dict if exp_filter[name]]


def select_last_record_of_experiments(user_range, exp_record_dict):
    experiments = select_experiments(user_range=user_range, exp_record_dict=exp_record_dict)
    records = [load_experiment(ex).get_latest_record(only_completed=True, err_if_none=False) for ex in experiments]
    if None in records:
        print('WARNING: Experiments {} have no completed records.', [e for e, r in izip_equal(experiments, records) if r is None])
    return records


def _filter_experiments(user_range, exp_record_dict):

    if user_range in exp_record_dict:
        is_in = [k==user_range for k in exp_record_dict]
    else:
        number_range = interpret_numbers(user_range)
        if number_range is not None:
            # experiment_ids = [experiment_list[i] for i in number_range]
            is_in = [i in number_range for i in xrange(len(exp_record_dict))]
        elif user_range == 'all':
            # experiment_ids = experiment_list
            is_in = [True]*len(exp_record_dict)
        elif user_range.startswith('has:'):
            phrase = user_range[len('has:'):]
            # experiment_ids = [exp_id for exp_id in experiment_list if phrase in exp_id]
            is_in = [phrase in exp_id for exp_id in exp_record_dict]
        elif user_range.startswith('1diff:'):
            # select experiments whose arguments differ by one element from the selected experiments
            base_range = user_range[len('1diff:'):]
            base_range_exps = select_experiments(base_range, exp_record_dict) # list<experiment_id>
            all_exp_args_hashes = {eid: set(compute_fixed_hash(a) for a in load_experiment(eid).get_args().items()) for eid in exp_record_dict} # dict<experiment_id : set<arg_hashes>>
            # assert all_equal_length(all_exp_args_hashes.values()), 'All variants must have the same number of arguments' # Note: we diable this because we may have lists of experiments with different root functions.
            is_in = [any(len(all_exp_args_hashes[eid].difference(all_exp_args_hashes[other_eid]))<=1 for other_eid in base_range_exps) for eid in exp_record_dict]
        elif user_range.startswith('hasnot:'):
            phrase = user_range[len('hasnot:'):]
            # experiment_ids = [exp_id for exp_id in experiment_list if phrase not in exp_id]
            is_in = [phrase not in exp_id for exp_id in exp_record_dict]
        elif user_range in ('unfinished', 'invalid', 'corrupt'):  # Return all experiments where all records are unfinished/invalid/corrupt
            record_filters = _filter_records(user_range, exp_record_dict)
            # experiment_ids = [exp_id for exp_id in experiment_list if len(record_filters[exp_id])]
            is_in = [all(record_filters[exp_id]) for exp_id in exp_record_dict]
        else:
            raise RecordSelectionError("Don't know how to use input '{}' to select experiments".format(user_range))

    return OrderedDict((exp_id, exp_is_in) for exp_id, exp_is_in in izip_equal(exp_record_dict, is_in))


def select_experiment_records(user_range, exp_record_dict, flat=True, load_records = True):
    """
    :param user_range:
    :param exp_record_dict: An OrderedDict<experiment_name: list<experiment_record_name>>
    :param flat: Return a list of experiment records, instead of an OrderedDict
    :return: if not flat, an An OrderedDict<experiment_name: list<experiment_record_name>>
        otherwise a list<experiment_record_name>
    """
    filters = _filter_records(user_range, exp_record_dict)
    filtered_dict = _select_record_from_filters(filters, exp_record_dict) if load_records else _select_record_ids_from_filters(filters, exp_record_dict)
    if flat:
        return [record_id for records in filtered_dict.values() for record_id in records]
    else:
        return filtered_dict


def _select_record_from_filters(filters, exp_record_dict):
    return OrderedDict((k, [load_experiment_record(record_id) for record_id, f in izip_equal(exp_record_dict[k], filters[k]) if f]) for k in exp_record_dict.keys())


def _select_record_ids_from_filters(filters, exp_record_dict):
    return OrderedDict((k, [record_id for record_id, f in izip_equal(exp_record_dict[k], filters[k]) if f]) for k in exp_record_dict.keys())


def _bitwise_and(a, b):
    return [a_ and b_ for a_, b_ in izip_equal(a, b)]


def _bitwise_or(a, b):
    return [a_ or b_ for a_, b_ in izip_equal(a, b)]


def _bitwise_andcascade(a, b):
    """
    :param a: A list of booleans whose length matches the number of True elements in b
    :param b: A list of booleans
    :return: A list
    """
    assert sum(b) == len(a), 'The number of elements in b: {}, did not match the number of true elements in a: {}'.format(sum(b), len(a))
    a_iter = iter(a)
    return [b_ and next(a_iter) for b_ in b]


def _bitwise_not(a):
    return [not a_ for a_ in a]


def _bitwise_filter_op(op, *filter_sets):

    output_set = filter_sets[0].copy()
    if op=='not':
        assert len(filter_sets)==1
        for k in output_set.keys():
            output_set[k] = _bitwise_not(filter_sets[0][k])
    elif op in ('and', 'or'):
        for k in output_set.keys():
            output_set[k] = reduce(_bitwise_and if op=='and' else _bitwise_or, [fs[k] for fs in filter_sets])
    elif op=='andcascade':
        for k in output_set.keys():
            output_set[k] = reduce(_bitwise_andcascade, [fs[k] for fs in filter_sets[::-1]])
    else:
        raise AssertionError('op should be one of {}'.format(('and', 'or', 'andcascade', 'not')))
    return output_set


_named_record_filters = {}
_named_record_filters['old'] = lambda rec_ids: ([True]*(len(rec_ids)-1)+[False]) if len(rec_ids)>0 else []
_named_record_filters['corrupt'] = lambda rec_ids: [load_experiment_record(rec_id).info.get_status_field()==ExpStatusOptions.CORRUPT for rec_id in rec_ids]
_named_record_filters['finished'] = lambda rec_ids: [load_experiment_record(rec_id).info.get_field(ExpInfoFields.STATUS) == ExpStatusOptions.FINISHED for rec_id in rec_ids]
_named_record_filters['invalid'] = lambda rec_ids: [load_experiment_record(rec_id).args_valid() is False for rec_id in rec_ids]
_named_record_filters['all'] = lambda rec_ids: [True]*len(rec_ids)
_named_record_filters['errors'] = lambda rec_ids: [load_experiment_record(rec_id).info.get_field(ExpInfoFields.STATUS)==ExpStatusOptions.ERROR for rec_id in rec_ids]
_named_record_filters['result'] = lambda rec_ids: [load_experiment_record(rec_id).has_result() for rec_id in rec_ids]
_named_record_filters['running'] = lambda rec_ids: [load_experiment_record(rec_id).info.get_field(ExpInfoFields.STATUS)==ExpStatusOptions.STARTED for rec_id in rec_ids]


def _filter_records(user_range, exp_record_dict):
    """
    :param user_range:
    :param exp_record_dict:
    :return: An OrderedDict<experiment_id -> list<True or False>> indicating whether each record from the given experiment passed the filter
    """

    if user_range=='unfinished':
        return _filter_records('~finished', exp_record_dict)
    elif user_range=='last':
        return _filter_records('~old', exp_record_dict)
    elif '|' in user_range:
        return _bitwise_filter_op('or', *[_filter_records(subrange, exp_record_dict) for subrange in user_range.split('|')])
    elif '&' in user_range:
        return _bitwise_filter_op('and', *[_filter_records(subrange, exp_record_dict) for subrange in user_range.split('&')])
    elif '@' in user_range:
        ix = user_range.index('@')
        first_part, second_part = user_range[:ix], user_range[ix+1:]
        _first_stage_filters = _filter_records(first_part, exp_record_dict)
        _new_dict = _select_record_ids_from_filters(_first_stage_filters, exp_record_dict)
        _second_stage_filters = _filter_records(second_part, _new_dict)
        return _bitwise_filter_op('andcascade', _first_stage_filters, _second_stage_filters)

    elif user_range.startswith('~'):
        return _bitwise_filter_op('not', _filter_records(user_range[1:], exp_record_dict))

    base = OrderedDict((k, [False]*len(v)) for k, v in exp_record_dict.items())
    if user_range in exp_record_dict:  # User just lists an experiment
        base[user_range] = [True]*len(base[user_range])
        return base

    number_range = interpret_numbers(user_range)
    keys = list(exp_record_dict.keys())

    if user_range in _named_record_filters:  # e.g. 'finished'
        for exp_id, _ in base.items():
            base[exp_id] = _named_record_filters[user_range](exp_record_dict[exp_id])
    elif number_range is not None:  # e.g. '6-12'
        for i in number_range:
            if i>len(keys):
                raise RecordSelectionError('Experiment {} does not exist (they go from 0 to {})'.format(i, len(keys)-1))
            base[keys[i]] = [True]*len(base[keys[i]])
    elif '.' in user_range:  # e.b. 6.3-4
        exp_rec_pairs = interpret_record_identifier(user_range)
        for exp_number, rec_number in exp_rec_pairs:
            if rec_number>=len(base[keys[exp_number]]):
                raise RecordSelectionError('Selection {}.{} does not exist.'.format(exp_number, rec_number))
            base[keys[exp_number]][rec_number] = True
    elif user_range.startswith('since:'):  # eg. 'since:24'
        time_id = user_range[len('since:'):]
        try:
            seconds_ago = int(time_id)*3600
        except:
            raise RecordSelectionError('Cannot interpret "{}" as a time.  Currently, it should be an integer, which means "within the last X hours"'.format(time_id))
        current_time = time()
        for exp_id, _ in base.items():
            base[exp_id] = [(current_time - load_experiment_record(rec_id).get_timestamp())<seconds_ago for rec_id in exp_record_dict[exp_id]]
    elif user_range.startswith('dur'):  # Eg dur<25  Means "All records that ran less than 25s"
        try:
            sign = user_range[3]
            assert sign in ('<', '>')
            seconds = int(user_range[4:])
        except:
            raise RecordSelectionError('Could not interpret "{}" as a duration.  Example is dur<25 to select all experiments that ran less than 25s.'.format(user_range))
        for exp_id, _ in base.items():
            durations = [load_experiment_record(rec_id).info.get_field(ExpInfoFields.RUNTIME, default = None) for rec_id in exp_record_dict[exp_id]]
            base[exp_id] = [False if dur is None else dur<seconds if sign=='<' else dur>seconds for dur in durations]
    else:
        raise RecordSelectionError("Don't know how to interpret subset '{}'.  Possible subsets: {}".format(user_range, list(_named_record_filters.keys())))
    return base


class RecordSelectionError(Exception):

    pass


def _filter_experiment_record_list(user_range, experiment_record_ids):
    if user_range=='all':
        return [True]*len(experiment_record_ids)
    elif user_range=='new':
        return detect_duplicates(experiment_record_ids, key=record_id_to_experiment_id, keep_last=True)
        # return [n for n, is_old in izip_equal(get_record_ids(), old) if not old]
    elif user_range=='old':
        return [not x for x in _filter_records(user_range, 'new')]
    elif user_range=='orphans':
        orphans = []
        global_lib = get_global_experiment_library()
        for i, record_id in enumerate(experiment_record_ids):
            info = load_experiment_record(record_id).info
            if 'Module' in info:
                try:
                    import_module(info['Module'])
                    if not record_id_to_experiment_id(record_id) in global_lib:
                        orphans.append(True)
                    else:
                        orphans.append(False)
                except ImportError:
                    orphans.append(True)
            else:  # They must be old... lets kill them!
                orphans.append(True)
        return orphans
    # elif user_range
    else:
        which_ones = interpret_numbers(user_range)
        if which_ones is None:
            raise Exception('Could not interpret user range: "{}"'.format(user_range))
        filters = [False]*len(experiment_record_ids)
        for i in which_ones:
            filters[i] = True
        return filters


def select_experiment_records_from_list(user_range, experiment_records):
    return [rec_id for rec_id, f in izip_equal(experiment_records, _filter_experiment_record_list(user_range, experiment_records)) if f]


def interpret_record_identifier(user_range):
    """
    You can identify a single record with, eg 3.4, meaning "record 4 from experiment 3:.
    You can identify a range with, eg 3.1-3, meaning "records 3.1, 3.2, 3.3"
    :param user_range: The user input
    :return: A list of 2 tuples (exp_no, record_no).  e.g [(3, 4)], or [(3, 1), (3, 2), (3, 3)] in the above examples.
    """
    if ',' in user_range:
        parts = user_range.split(',')
        return [pair for p in parts for pair in interpret_record_identifier(p)]
    if '.' not in user_range:
        raise RecordSelectionError('All record selections must have a "." - Your selection: "{}" did not.'.format(user_range))
    else:
        exp_number, record_numbers = user_range.split('.')
        return [(int(exp_number), rec_num) for rec_num in interpret_numbers(record_numbers)]


def interpret_numbers(user_range):
    """
    :param user_range: A string specifying a range of numbers.  Eg.
        interpret_numbers('4-6')==[4,5,6]
        interpret_numbers('4,6')==[4,6]
        interpret_numbers('4,6-9')==[4,6,7,8,9]
    :return: A list of integers, or None if the input is not numberic
    """
    if all(d in '0123456789-,' for d in user_range):
        numbers_and_ranges = user_range.split(',')
        numbers = [n for lst in [[int(s)] if '-' not in s else range(int(s[:s.index('-')]), int(s[s.index('-')+1:])+1) for s in numbers_and_ranges] for n in lst]
        return numbers
    else:
        return None

def run_experiment(experiment, slurm_job = False, experiment_path=None, **experiment_record_kwargs):
    """
    Run an experiment and save the results.  Return a string which uniquely identifies the experiment.
    You can run the experiment again later by calling show_experiment(location_string):

    :param experiment: The experiment object to be run
    :param slurm_job: It True, this function is interpreted as being run from within a SLURM call.
    :param experiment_path: If not None, the 'experiment_directory' option in the 'experiments' section of the .artemisrc file will be temporarily set to this value
    :param experiment_record_kwargs: Passed to ExperimentRecord.

    :return: A location_string, uniquely identifying the experiment.
    """
    if slurm_job:
        """
        If we run an experiment with slurm, then each subprocesses on the SLURM node will try to execute the experiment.
        This has two implications:
        1.) This run_experiment call is executed on a different processor than the processor on which the experiment UI is being executed. Consequently the environment is potentially different and 
        all global variables are reset to their default at artemis load.
        2.) Since we leave the distributed computation to the individual experiment to be executed, we catch these multiple executions here. Instead, we will only allow the first of the slurm nodes to proceed. All other nodes immediatly return.
        The fact that all other nodes immediately return does not impact the SLURM call since the SLURM call is considered finished only when all SLURM nodes terminate.
        I am aware that we could potentially save code and make this super slick by designing a subclass of Experiment which would be a 'DistributedSlurmExperiment', but this is future work.
        For now, this works.
        """
        assert "SLURM_NODEID" in os.environ.keys(), "You indicated that the experiment '{}' is run within a SLURM call, however the environment variable 'SLURM_NODEID' could not be found".format(experiment.get_id())
        if int(os.environ["SLURM_NODEID"]) > 0:
            return
    if experiment_path:
        'As mentioned above, global variables are reset, so I reset the one element I actually use' #TODO: Make this more elegant
        set_non_persistent_config_value(config_filename=".artemisrc", section="experiments", option="experiment_directory", value=experiment_path)

    return experiment.run(**experiment_record_kwargs)


def run_experiment_by_name(name, exp_dict='global', slurm_job=False, experiment_path=None, **experiment_record_kwargs):
    """
    Run an experiment and save the results.  Return a string which uniquely identifies the experiment.
    You can run the experiment again later by calling show_experiment(location_string):

    :param name: The name for the experiment (must reference something in exp_dict)
    :param exp_dict: A dict<str:func> where funcs is a function with no arguments that run the experiment.
    :param slurm_job: It True, this function is interpreted as being run from within a SLURM call.
    :param experiment_path: If not None, the 'experiment_directory' option in the 'experiments' section of the .artemisrc file will be temporarily set to this value
    :param experiment_record_kwargs: Passed to ExperimentRecord.

    :return: A location_string, uniquely identifying the experiment.
    """
    if exp_dict == 'global':
        exp_dict = get_global_experiment_library()
    experiment = exp_dict[name]
    return run_experiment(experiment,slurm_job, experiment_path, **experiment_record_kwargs)


def run_experiment_ignoring_errors(name, **kwargs):
    try:
        return run_experiment_by_name(name, **kwargs)
    except Exception as err:
        traceback.print_exc()


def run_multiple_experiments_with_slurm(experiments, n_parallel=None, max_processes_per_node=None, raise_exceptions=True, run_args={}, slurm_kwargs={}):
    '''
    Run multiple experiments using slurm, optionally in parallel.
    '''
    if n_parallel and n_parallel > 1:
        # raise NotImplementedError("No parallel Slurm execution at the moment. Implement it!")
        print ('Warning... parallel-slurm integration is very beta. Use with caution')
        experiment_subsets = divide_into_subsets(experiments, subset_size=n_parallel)
        for i, exp_subset in enumerate(experiment_subsets):
            nanny = Nanny()
            function_call = partial(run_multiple_experiments,
                experiments=exp_subset,
                parallel=n_parallel if max_processes_per_node is None else max_processes_per_node,
                display_results=False,
                run_args = run_args
                )
            spp = SlurmPythonProcess(name="Group %i"%i, function=function_call,ip_address="127.0.0.1", slurm_kwargs=slurm_kwargs)
            # Using Nanny only for convenient stdout & stderr forwarding.
            nanny.register_child_process(spp,monitor_for_termination=False)
            nanny.execute_all_child_processes(time_out=2)
    else:
        for i,exp in enumerate(experiments):
            nanny = Nanny()
            function_call = partial(run_experiment, experiment=exp, slurm_job=True, experiment_path=get_experiment_dir(),
                raise_exceptions=raise_exceptions,display_results=False, **run_args)
            spp = SlurmPythonProcess(name="Exp %i"%i, function=function_call,ip_address="127.0.0.1", slurm_kwargs=slurm_kwargs)
            # Using Nanny only for convenient stdout & stderr forwarding.
            nanny.register_child_process(spp,monitor_for_termination=False)
            nanny.execute_all_child_processes(time_out=2)


def _parallel_run_target(experiment_id_and_prefix, raise_exceptions, **kwargs):
    experiment_id, prefix = experiment_id_and_prefix
    if raise_exceptions:
        return run_experiment_by_name(experiment_id, prefix=prefix, **kwargs)
    else:
        return run_experiment_ignoring_errors(experiment_id, prefix=prefix, **kwargs)


def run_multiple_experiments(experiments, prefixes = None, parallel = False, display_results=False, raise_exceptions=True, notes = (), run_args = {}):
    """
    Run multiple experiments, optionally in parallel with multiprocessing.

    :param experiments: A collection of experiments
    :param parallel: Can be:
        True/'all': Run in parallel with as many processes as CPUs
        An integer indicating the number of processes to run
        False/None Don't run in parallel.
    :param raise_exceptions: Terminate exectution when one experiment fails.
    :param run_args: Other args to pass to Experiment.run()
    :return: A collection of experiment records.
    """

    if parallel:
        if parallel in (True, 'all'):
            parallel = multiprocessing.cpu_count()
        else:
            assert isinstance(parallel, int)
        experiment_identifiers = [ex.get_id() for ex in experiments]
        if prefixes is None:
            prefixes = range(len(experiment_identifiers))
        prefixes = [s+': ' for s in equalize_string_lengths(prefixes, side='right')]
        print ('Prefix key: \n'+'\n'.join('{}{}'.format(p, eid) for p, eid in izip_equal(prefixes, experiment_identifiers)))
        target_func = partial(_parallel_run_target, notes=notes, raise_exceptions=raise_exceptions, **run_args)
        p = multiprocessing.Pool(processes=parallel)

        return p.map(target_func, zip(experiment_identifiers, prefixes))
    else:
        return [ex.run(raise_exceptions=raise_exceptions, display_results=display_results, notes=notes, **run_args) for ex in experiments]


def remove_common_results_prefix(results_dict):
    """
    Remove the common prefix for the results you are comparing.
    :param results_dict: An OrderedDict of experiment Results
    :return: An OrderedDict of results with the common beginnings of the keys truncated.
    """
    # TODO: Fix this so that it splits correctly, not just on '.', which is not necessarily a separator.
    assert isinstance(dict, OrderedDict), 'Expecting an OrderedDict of <experiment_name -> result>'

    split_keys = [k.split('.') for k in results_dict.keys()]
    trimmed_keys = remove_common_prefix(split_keys)
    return OrderedDict((k, v) for k, v in izip_equal(trimmed_keys, results_dict.values()))


def get_experient_to_record_dict(experiment_ids = None):
    """
    Given a list of experiment ids, return an OrderedDict whose keys are the experiment ids and whose values
    are lists of experiment record ids.

    :param experiment_ids: A list of experiment ids.  (Defaults to all imported experiments)
    :return: A dict<experiment_id -> list<experiment_record_id>
    """
    if experiment_ids is None:
        experiment_ids = get_global_experiment_library().keys()
    record_ids = get_all_record_ids(experiment_ids)
    exp_rec_dict = OrderedDict((exp_id, []) for exp_id in experiment_ids)
    for rid in record_ids:
        rec = load_experiment_record(rid)
        exp_id = rec.get_experiment_id()
        exp_rec_dict[exp_id].append(rid)
    return exp_rec_dict


def deprefix_experiment_ids(experiment_ids):
    """
    Given a list of experiment ids, removed the common root experiments from the list.
    :param experiment_ids: A list of experiment ids.
    :return: A list of experiment ids with the root prefix removed.
    """

    # First build dict mapping experiment_ids to their parents experiment_ids
    glib = get_global_experiment_library()
    exp_to_parent = {}
    for eid in glib.keys():
        ex = glib[eid]
        for var in ex.get_variants():
            exp_to_parent[var.get_id()] = ex.get_id()

    @memoize
    def get_experiment_tuple(exp_id):
        if exp_id in exp_to_parent:
            parent_id = exp_to_parent[exp_id]
            parent_tuple = get_experiment_tuple(parent_id)
            return parent_tuple + (exp_id[len(parent_id)+1:], )
        else:
            return (exp_id, )

    # Then for each experiment in the list,
    tuples = [get_experiment_tuple(eid) for eid in experiment_ids]
    de_prefixed_tuples = remove_common_prefix(tuples, keep_base=False)
    start_with = '' if len(de_prefixed_tuples[0])==len(tuples[0]) else '.'
    new_strings = [start_with+'.'.join(ex_tup) for ex_tup in de_prefixed_tuples]
    return new_strings



from functools import partial
import time

import numpy as np
from artemis.plotting.demo_dbplot import demo_dbplot
from artemis.plotting.db_plotting import dbplot, clear_dbplot, hold_dbplots, freeze_all_dbplots, reset_dbplot, \
    dbplot_hang
from artemis.plotting.matplotlib_backend import LinePlot, HistogramPlot, MovingPointPlot, is_server_plotting_on
import pytest
from matplotlib import pyplot as plt
from matplotlib import gridspec
from six.moves import xrange

__author__ = 'peter'


def test_dbplot(n_steps = 3):

    reset_dbplot()

    arr = np.random.rand(10, 10)
    for i in xrange(n_steps):
        arr_sq=arr**2
        arr = arr_sq/np.mean(arr_sq)
        dbplot(arr, 'arr')
        for j in xrange(3):
            barr = np.random.randn(10, 2)
            dbplot(barr, 'barr', plot_type=partial(LinePlot))


@pytest.mark.skipif('True', reason = 'Need to make matplotlib backend work with scales.')
def test_dbplot_logscale(n_steps = 3):
    reset_dbplot()

    arr = np.random.rand(10, 10)

    for i in xrange(n_steps):
        arr_sq=arr**2
        arr = arr_sq/np.mean(arr_sq)
        dbplot(arr, 'arr')
        for j in xrange(3):
            barr = np.random.randn(10, 2)
            kw = {"y_axis_type":"log"}
            dbplot(barr, 'barr', plot_type=partial(LinePlot,y_axis_type='log'))


def test_particular_plot(n_steps = 3):
    reset_dbplot()

    for i in xrange(n_steps):
        r = np.random.randn(1)
        dbplot(r, plot_type=partial(HistogramPlot,edges=np.linspace(-5, 5, 20)))


def test_history_plot_updating():
    """
    This test checks that we've fixed the bug mentioned in issue 1: https://github.com/QUVA-Lab/artemis/issues/1
    That was, when you are updating multiple plots with history in a loop, everytime any of the plots is updated, they
    all get updated with the most recent data.  You'll see this in plot 'c' - with the bug, it moves in steps, with 3
    of the same sample in a row.  If it works it should be spikey.
    """
    reset_dbplot()
    for i in xrange(10):
        dbplot(np.random.randn(20, 20), 'a')
        dbplot(np.random.randn(20, 20), 'b')
        dbplot(np.random.randn(), 'c', plot_type=partial(MovingPointPlot))
        # dbplot(np.random.randn(), 'c', plot_type=partial(MovingPointPlot, memory=2))


def test_moving_point_multiple_points():
    reset_dbplot()
    for i in xrange(5):
        dbplot(np.sin([i/10., i/15.]), 'unlim buffer', plot_type = partial(MovingPointPlot))
        dbplot(np.sin([i/10., i/15.]), 'lim buffer', plot_type = partial(MovingPointPlot,buffer_len=20))

def test_same_object():
    """
    There was a bug where when you plotted two of the same array, you got "already seen object".  This tests makes
    sure it's gotten rid of.  If it's gone, both matrices should plot.  Otherwise you'll get "Already seen object" showing
    up on one of the plots.
    """
    reset_dbplot()
    a = np.random.randn(20, 20)
    for _ in xrange(5):
        dbplot(a, 'a')
        dbplot(a, 'b')


def test_multiple_figures():
    reset_dbplot()
    for _ in xrange(2):
        dbplot(np.random.randn(20, 20), 'a', fig='1')
        dbplot(np.random.randn(20, 20), 'b', fig='1')
        dbplot(np.random.randn(20, 20), 'c', fig='2')
        dbplot(np.random.randn(20, 20), 'd', fig='2')


def test_list_of_images():
    reset_dbplot()
    for _ in xrange(2):
        dbplot([np.random.randn(12, 30), np.random.randn(10, 10), np.random.randn(15, 10)])


def test_two_plots_in_the_same_axis_version_1():
    reset_dbplot()
    # Option 1: Name the 'axis' argument to the second plot after the name of the first
    for i in xrange(5):
        data = np.random.randn(200)
        x = np.linspace(-5, 5, 100)
        with hold_dbplots():
            dbplot(data, 'histogram', plot_type='histogram')
            dbplot((x, 1./np.sqrt(2*np.pi*np.var(data)) * np.exp(-(x-np.mean(data))**2/(2*np.var(data)))), 'density', axis='histogram', plot_type='line')


def test_two_plots_in_the_same_axis_version_2():
    reset_dbplot()
    # Option 2: Give both plots the same 'axis' argument
    for i in xrange(5):
        data = np.random.randn(200)
        x = np.linspace(-5, 5, 100)
        with hold_dbplots():
            dbplot(data, 'histogram', plot_type='histogram', axis='hist')
            dbplot((x, 1./np.sqrt(2*np.pi*np.var(data)) * np.exp(-(x-np.mean(data))**2/(2*np.var(data)))), 'density', axis='hist', plot_type='line')


@pytest.mark.skipif(is_server_plotting_on(), reason = "This fails in server mode because we currently do not have an interpretation of freeze_all_dbplots")
def test_freeze_dbplot():
    reset_dbplot()
    def random_walk():
        data = 0
        for i in xrange(10):
            data += np.random.randn()
            dbplot(data, 'walk')#, plot_type=lambda: MovingPointPlot(axes_update_mode='expand'))

    random_walk()
    freeze_all_dbplots()
    random_walk()


def test_trajectory_plot():

    for i in xrange(5):
        dbplot((np.cos(i/10.), np.sin(i/11.)), 'path', plot_type='trajectory')


def test_demo_dbplot():
    demo_dbplot(n_frames=3)
    clear_dbplot()


def test_custom_axes_placement(hang=False):

    gs1 = gridspec.GridSpec(3, 1, left=0, right=0.5, hspace=0)
    dbplot(np.sin(np.linspace(0, 10, 100)), 'a', plot_type='line', axis=gs1[0, 0])
    dbplot(np.sin(np.linspace(0, 10, 100)+1), 'b', plot_type='line', axis=gs1[1, 0])
    dbplot(np.sin(np.linspace(0, 10, 100)+2), 'c', plot_type='line', axis=gs1[2, 0])

    gs2 = gridspec.GridSpec(2, 1, left=0.5, right=1, hspace=0.1)
    dbplot(np.random.randn(20, 20), 'im1', axis=gs2[0, 0])
    dbplot(np.random.randn(20, 20, 3), 'im2', axis=gs2[1, 0])

    if hang:
        dbplot_hang()


def test_cornertext():

    dbplot(np.random.randn(5, 5), 'a', cornertext='one')
    dbplot(np.random.randn(5, 5), 'a', cornertext='two')
    dbplot(np.random.randn(5, 5), 'a', cornertext='three')


def test_close_and_open():

    for _ in xrange(20):
        dbplot(np.random.randn(5), 'a')

    plt.close(plt.gcf())

    for _ in xrange(20):
        dbplot(np.random.randn(5), 'b')


if __name__ == '__main__':
    if is_server_plotting_on():
        test_cornertext()
        time.sleep(2.)
        test_trajectory_plot()
        time.sleep(2.)
        test_demo_dbplot()
        time.sleep(2.)
        test_two_plots_in_the_same_axis_version_1()
        time.sleep(2.)
        test_two_plots_in_the_same_axis_version_2()
        time.sleep(2.)
        test_moving_point_multiple_points()
        time.sleep(2.)
        test_list_of_images()
        time.sleep(2.)
        test_multiple_figures()
        time.sleep(2.)
        test_same_object()
        time.sleep(2.)
        test_history_plot_updating()
        time.sleep(2.)
        test_particular_plot()
        time.sleep(2.)
        test_dbplot()
        time.sleep(2.)
        test_custom_axes_placement()
        time.sleep(2.)
        test_close_and_open()
        time.sleep(2.)
    else:
        test_cornertext()
        test_trajectory_plot()
        test_demo_dbplot()
        test_freeze_dbplot()
        test_two_plots_in_the_same_axis_version_1()
        test_two_plots_in_the_same_axis_version_2()
        test_moving_point_multiple_points()
        test_list_of_images()
        test_multiple_figures()
        test_same_object()
        test_history_plot_updating()
        test_particular_plot()
        test_dbplot()
        test_custom_axes_placement()
        test_close_and_open()



import itertools
import time
import warnings
import os
from collections import OrderedDict

import matplotlib.pyplot as plt
import numpy as np
import pickle
import pytest
from six.moves import xrange

from artemis.experiments.decorators import experiment_function, experiment_root
from artemis.experiments.deprecated import start_experiment, end_current_experiment
from artemis.experiments.experiment_management import run_multiple_experiments
from artemis.experiments.experiment_record import \
    load_experiment_record, ExperimentRecord, record_experiment, \
    delete_experiment_with_id, get_current_record_dir, open_in_record_dir, \
    ExpStatusOptions, clear_experiment_records, get_current_experiment_id, get_current_experiment_record, \
    get_current_record_id, has_experiment_record, experiment_id_to_record_ids
from artemis.experiments.experiments import get_experiment_info, load_experiment, experiment_testing_context, \
    clear_all_experiments
from artemis.general.test_mode import set_test_mode

__author__ = 'peter'

"""
The experiment interface.  Currently, this can be used in many ways.  We'd like to
converge on just using the interface demonstrated in test_experiment_api.  So
use that if you're looking for an example.
"""


@experiment_function
def experiment_test_function(seed=1234):

    rng = np.random.RandomState(seed)
    with warnings.catch_warnings():  # This is just to stop that stupid matplotlib warning from screwing up our logs.
        warnings.simplefilter('ignore')

        print('aaa')
        plt.figure('sensible defaults')
        dat = rng.randn(4, 5)
        plt.subplot(211)
        plt.imshow(dat)
        plt.subplot(212)
        plt.imshow(dat, interpolation = 'nearest', cmap = 'gray')
        plt.show()
        print('bbb')
        plt.figure()
        plt.plot(rng.randn(10))
        plt.show()


def assert_experiment_record_is_correct(exp_rec, show_figures=False):

    assert exp_rec.get_log() == 'aaa\nbbb\n'
    assert len(exp_rec.get_figure_locs()) == 2

    # Now assert that you can load an experiment from file and again display the figures.
    exp_dir = exp_rec.get_dir()
    exp_rec_copy = ExperimentRecord(exp_dir)
    assert exp_rec_copy.get_log() == 'aaa\nbbb\n'
    if show_figures:
        exp_rec_copy.show_figures()
    assert len(exp_rec_copy.get_figure_locs()) == 2


def test_experiment_with():
    """
    DEPRECATED INTERFACE

    This syntax uses the record_experiment function directly instead of hiding it.
    """

    with experiment_testing_context():
        delete_experiment_with_id('test_exp')
        with record_experiment(identifier = 'test_exp', print_to_console=True) as exp_rec:
            experiment_test_function()
            assert_experiment_record_is_correct(exp_rec, show_figures=False)


def test_start_experiment():
    """
    DEPRECATED INTERFACE

    An alternative syntax to the with statement - less tidy but possibly better
    for notebooks and such because it avoids you having to indent all code in the
    experiment.
    """

    delete_experiment_with_id('start_stop_test')
    with experiment_testing_context():
        record = start_experiment('start_stop_test')
        experiment_test_function()
        end_current_experiment()
        assert_experiment_record_is_correct(record, show_figures=False)


def test_get_latest():
    with experiment_testing_context():
        record_1 = experiment_test_function.run()
        time.sleep(0.01)
        record_2 = experiment_test_function.run()
        identifier = load_experiment('experiment_test_function').get_latest_record().get_id()
        assert identifier == record_2.get_id()


def test_get_latest_identifier():

    with experiment_testing_context():
        exp_rec = experiment_test_function.run()
        print(get_experiment_info('experiment_test_function'))
        assert_experiment_record_is_correct(exp_rec)
        last_experiment_identifier = load_experiment('experiment_test_function').get_latest_record().get_id()
        assert last_experiment_identifier is not None, 'Experiment was run, this should not be none'
        same_exp_rec = load_experiment_record(last_experiment_identifier)
        assert_experiment_record_is_correct(same_exp_rec)


def test_accessing_experiment_dir():

    with experiment_testing_context():

        @experiment_function
        def access_dir_test():
            print('123')
            print('abc')
            dir = get_current_record_dir()
            with open_in_record_dir('my_test_file.txt', 'w') as f:
                f.write('Experiment Directory is: {}'.format(dir))

        record = access_dir_test.run()

        filepaths = record.list_files()

        assert 'my_test_file.txt' in filepaths

        with record.open_file('my_test_file.txt') as f:
            assert f.read() == 'Experiment Directory is: {}'.format(record.get_dir())

        with record.open_file('output.txt') as f:
            assert f.read() == '123\nabc\n'


@experiment_function
def add_some_numbers_test_experiment(a=1, b=1):
    c = a + b
    print(c)
    return c


def test_saving_result():
    # Run root experiment
    with experiment_testing_context():
        rec = add_some_numbers_test_experiment.run()
        assert rec.get_result() == 2


def test_variants():

    @experiment_function
    def add_some_numbers(a=1, b=1):
        c=a+b
        print(c)
        return c

    with experiment_testing_context():

        # Create a named variant
        e1=add_some_numbers.add_variant('b is 3', b=3)
        assert e1.run().get_result()==4

        # Creata a sub-variant
        e11 = e1.add_variant('a is 2', a=2)
        assert e11.run().get_result() == 5

        # Create unnamed variant
        e2=add_some_numbers.add_variant(b=4)
        assert e2.run().get_result()==5
        assert e2.get_id() == 'add_some_numbers.b=4'

        # Create array of variants
        e_list = [add_some_numbers.add_variant(b=i) for i in xrange(5, 8)]
        assert [e.get_id() for e in e_list] == ['add_some_numbers.b=5', 'add_some_numbers.b=6', 'add_some_numbers.b=7']
        assert [e.run().get_result()==j for e, j in zip(e_list, range(6, 11))]

        # Create grid of variants
        e_grid = [add_some_numbers.add_variant(a=a, b=b) for a, b in itertools.product([2, 3], [4, 5, 6])]
        assert [e.get_id() for e in e_grid] == ['add_some_numbers.a=2,b=4', 'add_some_numbers.a=2,b=5', 'add_some_numbers.a=2,b=6',
                                                  'add_some_numbers.a=3,b=4', 'add_some_numbers.a=3,b=5', 'add_some_numbers.a=3,b=6']
        assert add_some_numbers.get_variant(a=2, b=4).run().get_result()==6
        assert add_some_numbers.get_variant(a=3, b=5).run().get_result()==8

        experiments = add_some_numbers.get_all_variants(include_roots=True, include_self=True)
        assert len(experiments)==13


@experiment_function
def my_api_test(a=1, b=3):
    print('aaa')
    return a*b

my_api_test.add_variant('a2b2', a=2, b=2)
my_api_test.add_variant('a3b2', a=3, b=2)


def test_experiment_api(try_browse=False):

    with experiment_testing_context():
        my_api_test.get_variant('a2b2').run()
        record = my_api_test.get_variant('a2b2').get_latest_record()

        assert record.get_log() == 'aaa\n'
        assert record.get_result() == 4
        assert record.get_args() == OrderedDict([('a', 2), ('b', 2)])
        assert record.get_status() == ExpStatusOptions.FINISHED

    if try_browse:
        my_api_test.browse()


def test_figure_saving(show_them = False):

    with experiment_testing_context():
        record = experiment_test_function.run()

    plt.close('all')  # Close all figures
    figs = record.load_figures()
    assert len(figs)==2
    if show_them:
        plt.show()


def test_get_variant_records_and_delete():

    with experiment_testing_context():

        for record in my_api_test.get_variant_records(flat=True):
            record.delete()

        assert len(my_api_test.get_variant_records(flat=True))==0

        my_api_test.run()
        my_api_test.get_variant('a2b2').run()

        assert len(my_api_test.get_variant_records(flat=True))==2

        for record in my_api_test.get_variant_records(flat=True):
            record.delete()

        assert len(my_api_test.get_variant_records(flat=True))==0


def test_experiments_play_well_with_debug():

    with experiment_testing_context():

        @experiment_function
        def my_simple_test():
            plt.show._needmain=False  # pyplot does this internally whenever a breakpoint is reached.
            return 1

        my_simple_test.run()


def test_run_multiple_experiments():

    with experiment_testing_context():

        experiments = my_api_test.get_all_variants()
        assert len(experiments)==3

        records = run_multiple_experiments(experiments)
        assert [record.get_result() for record in records] == [3, 4, 6]

        records = run_multiple_experiments(experiments, parallel=True)
        assert [record.get_result() for record in records] == [3, 4, 6]


def test_parallel_run_errors():

    with experiment_testing_context():

        @experiment_function
        def my_error_causing_test(a=1):
            raise Exception('nononono')

        my_error_causing_test.add_variant(a=2)

        variants = my_error_causing_test.get_all_variants()

        assert len(variants)==2

        run_multiple_experiments(variants, parallel=True, raise_exceptions=False)

        with pytest.raises(Exception) as err:
            run_multiple_experiments(variants, parallel=True, raise_exceptions=True)
        print("^^^ Dont't worry, the above is not actually an error, we were just asserting that we caught the error.")

        assert str(err.value) == 'nononono'


def test_invalid_arg_detection():
    """
    Check that we notice when an experiment is redefined with new args.
    """

    with experiment_testing_context(new_experiment_lib=True):

        @experiment_function
        def my_experiment_gfdsbhtds(a=1, b=[2, 3.], c={'a': 5, 'b': [6, 7]}):
            return a+1

        rec = my_experiment_gfdsbhtds.run()

        assert rec.args_valid()
        clear_all_experiments()

        @experiment_function
        def my_experiment_gfdsbhtds(a=1, b=[2, 3.], c={'a': 5, 'b': [6, 7]}):
            return a+1

        assert rec.args_valid()  # Assert that the args still match
        clear_all_experiments()

        @experiment_function
        def my_experiment_gfdsbhtds(a=1, b=[2, 3.], c={'a': 5, 'b': [6, 8]}):  # CHANGE IN ARGS!
            return a+1

        assert not rec.args_valid()


def test_invalid_arg_detection_2():
    """
    Check that we notice when an experiment is redefined with new args.
    """

    with experiment_testing_context(new_experiment_lib=True):

        a = {"a%s"%i for i in range(100)}

        @experiment_function
        def my_experiment_gfdsbhtds(a=a):
            return None

        rec = my_experiment_gfdsbhtds.run()

        assert rec.args_valid() is True
        clear_all_experiments()

        @experiment_function
        def my_experiment_gfdsbhtds(a=a):
            return None

        assert rec.args_valid() is True  # Assert that the args still match


def test_experiment_errors():

    with experiment_testing_context(new_experiment_lib=True):

        class MyManualException(Exception):
            pass

        @experiment_function
        def my_experiment_fdsgbdn():
            raise MyManualException()

        with pytest.raises(MyManualException):
            my_experiment_fdsgbdn.run()
        with pytest.raises(MyManualException):
            my_experiment_fdsgbdn.run()

        assert my_experiment_fdsgbdn.get_latest_record().get_status() == ExpStatusOptions.ERROR

        # Previously this caused an error because simple context managers didn't catch errors


def test_experiment_corrupt_detection():

    with experiment_testing_context(new_experiment_lib=True):

        @experiment_function
        def my_experiment_bfdsssdvs(a=1):
            return a+2

        r = my_experiment_bfdsssdvs.run()
        assert r.get_status() == ExpStatusOptions.FINISHED
        os.remove(os.path.join(r.get_dir(), 'info.pkl'))  # Manually remove the info file
        r = my_experiment_bfdsssdvs.get_latest_record()
        assert r.get_status() == ExpStatusOptions.CORRUPT


def test_current_experiment_access_functions():

    with experiment_testing_context(new_experiment_lib=True):

        @experiment_function
        def my_experiment_dfgsdgfdaf(a=1):

            assert a==4, 'Only meant to run aaa variant.'
            experiment_id = get_current_experiment_id()

            rec = get_current_experiment_record()

            record_id = get_current_record_id()

            assert record_id.endswith('-'+experiment_id)

            assert experiment_id == 'my_experiment_dfgsdgfdaf.aaa'
            loc = get_current_record_dir()

            _, record_dir = os.path.split(loc)

            assert record_dir == record_id
            assert os.path.isdir(loc)
            assert loc.endswith(record_id)

            with open_in_record_dir('somefile.pkl', 'wb') as f:
                pickle.dump([1, 2, 3], f, protocol=pickle.HIGHEST_PROTOCOL)

            assert os.path.exists(os.path.join(loc, 'somefile.pkl'))

            with open_in_record_dir('somefile.pkl', 'rb') as f:
                assert pickle.load(f) == [1, 2, 3]

            exp = rec.get_experiment()
            assert exp.get_id() == experiment_id
            assert exp.get_args() == rec.get_args() == OrderedDict([('a', 4)])
            assert rec.get_dir() == loc
            assert has_experiment_record(experiment_id)
            assert record_id in experiment_id_to_record_ids(experiment_id)
            return a+2

        v = my_experiment_dfgsdgfdaf.add_variant('aaa', a=4)

        v.run()


def test_generator_experiment():

    with experiment_testing_context(new_experiment_lib=True):
        @experiment_root
        def my_generator_exp(n_steps, poison_4 = False):
            for i in range(n_steps):
                if poison_4 and i==4:
                    raise Exception('Unlucky Number!')
                yield i

        X1 = my_generator_exp.add_variant(n_steps=5)
        X2 = my_generator_exp.add_variant(n_steps=5, poison_4 = True)

        rec1 = X1.run()
        rec2 = X2.run(raise_exceptions = False)

        assert rec1.get_result() == 4
        assert rec2.get_result() == 3


if __name__ == '__main__':

    set_test_mode(True)
    test_get_latest_identifier()
    test_get_latest()
    test_experiment_with()
    test_start_experiment()
    test_accessing_experiment_dir()
    test_saving_result()
    test_variants()
    test_experiment_api(try_browse=False)
    test_figure_saving(show_them=False)
    test_get_variant_records_and_delete()
    test_experiments_play_well_with_debug()
    test_run_multiple_experiments()
    test_parallel_run_errors()
    test_invalid_arg_detection()
    test_invalid_arg_detection_2()
    test_experiment_errors()
    test_experiment_corrupt_detection()
    test_current_experiment_access_functions()
    test_generator_experiment()



import os
import pickle
import numpy as np
from artemis.fileman.file_getter import get_file, get_archive
from artemis.ml.datasets.datasets import DataSet, DataCollection
from six.moves import xrange, input

__author__ = 'peter'


def get_cifar_100_dataset(n_training_samples=None, n_test_samples=None, whiten_individual_images=False, fine_labels=True):
    '''
    :param n_training_samples: Number of training samples, or None to leave it at 50000
    :param n_test_samples: Number of test samples, or None to leave it at 10000
    :param whiten_individual_images: True to normalize each individual image, and turn them from uint8 to double
    :param fine_labels: True to use the fine categorization (100 classes) or False to use the coarse categorization (20 classes)
    :return: The CIFAR-100 dataset, which consists of 50000 training and 10000 test images.
        Images are 32x32 uint8 (or double) RGB images (n_samples, 3, 32, 32) of 100 (or 20) categories of objects.
        Targets are integer labels in the range [0, 100] (or [0,20])

    '''

    directory = get_archive(relative_path='data/cifar-100', url='https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz')
    with open(os.path.join(directory, "cifar-100-python", "train"), 'rb') as fo:
        dict = pickle.load(fo)
    x_tr = dict["data"].reshape(-1, 3, 32, 32)
    y_tr = dict["fine_labels"] if fine_labels else dict["coarse_labels"]
    y_tr = np.array(y_tr)

    with open(os.path.join(directory, "cifar-100-python", "test"), 'rb') as fo:
        dict = pickle.load(fo)
    x_ts = dict["data"].reshape(-1, 3, 32, 32)
    y_ts = dict["fine_labels"] if fine_labels else dict["coarse_labels"]
    y_ts = np.array(y_ts)

    if whiten_individual_images:
        mean_tr = x_tr.mean(axis=(1,2,3), keepdims=True)
        std_tr = x_tr.std(axis=(1,2,3), keepdims=True)
        x_tr = (x_tr - mean_tr)/std_tr
        mean_ts = x_ts.mean(axis=(1,2,3), keepdims=True)
        std_ts = x_ts.std(axis=(1,2,3), keepdims=True)
        x_ts = (x_ts - mean_ts)/std_ts

    if n_training_samples is not None:
        x_tr = x_tr[:n_training_samples]
        y_tr = y_tr[:n_training_samples]
    if n_test_samples is not None:
        x_ts = x_ts[:n_test_samples]
        y_ts = y_ts[:n_test_samples]


    return DataSet(training_set=DataCollection(x_tr, y_tr), test_set=DataCollection(x_ts, y_ts), name='CIFAR-100-%s'%("fine" if fine_labels else "coarse"))






def get_cifar_10_dataset(n_training_samples = None, n_test_samples = None, normalize_inputs = False):
    """
    :param n_training_samples: Number of training samples, or None to leave it at 50000
    :param n_test_samples: Number of test samples, or None to leave it at 10000
    :param normalize_inputs: True to normalize inputs, and turn them from uint8 to double
    :param swap_axes: True to arrange images as (n_samples, n_colors, n_rows, n_cols) instead of (n_samples, n_rows, n_cols, n_colors)

    :return: The CIFAR-10 dataset, which consists of 50000 training and 10000 test images.
        Images are 32x32 uint8 RGB images (n_samples, 3, 32, 32) of 10 categories of objects.
        Targets are integer labels in the range [0, 9]
    """
    # TODO: Make method for downloading/unpacking data (from http://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz)
    # We have this for single files already, but in this case the gz contains a folder with the files in it.

    directory = get_archive(relative_path='data/cifar-10', url = 'http://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz')

    n_batches_to_read = 5 if n_training_samples is None else int(np.ceil(n_training_samples/10000.))

    file_paths = [get_file(os.path.join(directory, 'cifar-10-batches-py', 'data_batch_%s' % (i, ))) for i in xrange(1, n_batches_to_read+1)] \
        + [get_file(os.path.join(directory, 'cifar-10-batches-py', 'test_batch'))]

    data = []
    for file_path in file_paths:
        with open(file_path) as f:
            batch_data = pickle.load(f)
            data.append(batch_data)

    x_tr = np.concatenate([d['data'] for d in data[:-1]], axis = 0).reshape(-1, 3, 32, 32)
    y_tr = np.concatenate([d['labels'] for d in data[:-1]], axis = 0)
    x_ts = data[-1]['data'].reshape(-1, 3, 32, 32)
    y_ts = np.array(data[-1]['labels'])

    if normalize_inputs:
        mean = x_tr.mean(axis=0, keepdims=True)
        std = x_tr.std(axis=0, keepdims=True)
        x_tr = (x_tr - mean)/std
        x_ts = (x_ts - mean)/std

    if n_training_samples is not None:
        x_tr = x_tr[:n_training_samples]
        y_tr = y_tr[:n_training_samples]
    if n_test_samples is not None:
        x_ts = x_ts[:n_test_samples]
        y_ts = y_ts[:n_test_samples]

    return DataSet(training_set=DataCollection(x_tr, y_tr), test_set=DataCollection(x_ts, y_ts), name = 'CIFAR-10')


def get_cifar_10_labels():
    return [
        'airplane',
        'automobile',
        'bird',
        'cat',
        'deer',
        'dog',
        'frog',
        'horse',
        'ship',
        'truck']


if __name__ == '__main__':

    from artemis.plotting.easy_plotting import ezplot

    dataset = get_cifar_10_dataset()
    dataset = get_cifar_100_dataset()
    n_samples = 100

    ezplot({
        'sampled training images': np.rollaxis(dataset.training_set.input[:n_samples], 1, 4),# np.swapaxes(dataset.training_set.input[:n_samples], 1, 3).reshape(10, 10, 32, 32, 3),
        'sampled training labels': dataset.training_set.target[:n_samples].reshape(10, 10)
        }, cmap = 'jet')
    input("Enter to exit")



import os
import sys
from six.moves.configparser import NoSectionError, NoOptionError, ConfigParser

__author__ = 'peter'


_CONFIG_OBJECTS = {}


def get_config_value(config_filename, section, option, default_generator=None, write_default=False, read_method=None, use_cashed_config=True):
    """
    Get a setting from a configuration file.  If none exists, you can optionally create one.  An example config file is
    the ~/.theanorc file, which may contain:

        [global]                    # <-- section
        floatX = float32            # <-- option (value is 'float32')
        device = gpu0

        [lib]
        cnmem = 1

        [cuda]
        root=/usr/local/cuda/

    :param config_filename: A configuration filename (eg '.theanorc').  Here we enforce the convention that files start
        with '.', and are stored in the user-home directory.
    :param section: Section in the config file (as referred by squared brackets: see example above)
    :param option: The option of interest (see above example)
    :param default_generator: A function that generates the property if it is not there, or just the default value if
        default_generator is not callable
    :param write_default: Set to true if property was not found and you want to write the default value into the file.
    :param read_method: The method to read your setting.
        If left at None (default) it just returns the string.
        If 'eval' it parses the setting into a python object
        If it is a function, it passes the value through the function before returning it.
    :param use_cashed_config: If set, will not read the config file from the file system but use the previously read and stored config file.
        Since the cashed config file might have been modified since reading it from disk, the returned value might be different from the value
        stored on the file on disk. In case write_default is set to True, the default value will be written to disk either way if no value has been found.
        If set to False, the original value will be returned without modifying the hashed version.
    :return: The value of the property of interest.
    """
    config_path = get_config_path(config_filename)
    default_used = False

    if write_default:
        assert default_generator is not None, "If you set write_default true, you must provide a function that can generate the default."

    if not os.path.exists(config_path):
        assert default_generator is not None, 'No config file "%s" exists, and you do not have any default value.' % (config_path, )
        value = default_generator()
        default_used = True
    else:
        config = _get_config_object(config_path, use_cashed_config=use_cashed_config)
        try:
            value = config.get(section, option)
        except (NoSectionError, NoOptionError) as err:
            if default_generator is None:
                raise
            else:
                value = default_generator() if callable(default_generator) else default_generator
                default_used = True

    if default_used and write_default:
        config = _get_config_object(config_path)
        if not config.has_section(section):
            config.add_section(section)
        config.set(section, option, value)
        with open(config_path, 'w') as f:
            config.write(f)

    if read_method == 'eval':
        value = eval(value)
    elif callable(read_method):
        value = read_method(value)
    return value


def _get_config_object(config_path, use_cashed_config=True):
    '''
    Returns a ConfigParser for the config file at the given path. If no file exists, an empty config file is created.
    :param config_path:
    :param use_cashed_config: If set to True, will return the previously created ConfigParser file (if previously created).
        If set to False, will re-read the config file from disk. If a ConfigParser was previously created, it will not be replaced!
    :return:
    '''
    if config_path not in _CONFIG_OBJECTS or not use_cashed_config:
        config = ConfigParser()
        if not os.path.exists(config_path):
            with open(config_path,'w') as f:
                config.write(f)
        else:
            config.read(config_path)
        if use_cashed_config:
            _CONFIG_OBJECTS[config_path] = config
    else:
        config = _CONFIG_OBJECTS[config_path]
    return config


def get_config_sections(config_filename):
    config_path = get_config_path(config_filename)
    sections = _get_config_object(config_path).sections()
    return sections


def get_home_dir():
    # This function exists because of some weirdness when running remotely.
    home_dir = os.path.expanduser('~')
    if os.path.exists(home_dir):
        return home_dir
    elif sys.platform=='linux2':
        home_dir = os.path.join('/home', os.getlogin())
    elif sys.platform=='darwin':
        home_dir==os.path.join('/Users', os.getlogin())
    assert os.path.exists(home_dir), 'The home directory "{}" seems not to exist.  This is odd, to say the least.'.format(home_dir)
    return home_dir

def get_config_path(config_filename):
    assert config_filename.startswith('.'), "We enforce the convention that configuration files must start with '.'"
    config_path = os.path.join(get_home_dir(), config_filename)
    return config_path

def set_non_persistent_config_value(config_filename, section, option, value):
    config_path = get_config_path(config_filename)
    config = _get_config_object(config_path)
    if not config.has_section(section):
        config.add_section(section)
    return config.set(section=section, option=option,value=value)


if __name__ == '__main__':
    config_path = get_config_path('.artemisrc')
    print('Contents of {}:\n-------------\n'.format(config_path))
    with open(config_path) as f:
        print(f.read())



import atexit
import inspect
from collections import OrderedDict
from contextlib import contextmanager
from functools import partial
from six import string_types
from artemis.experiments.experiment_record import ExpStatusOptions, experiment_id_to_record_ids, load_experiment_record, \
    get_all_record_ids, clear_experiment_records
from artemis.experiments.experiment_record import run_and_record
from artemis.general.functional import infer_derived_arg_values, get_partial_root


class Experiment(object):
    """
    An experiment.  In general you should not use this class directly.  Use the experiment_function decorator, and
    create variants using decorated_function.add_variant()
    """

    def __init__(self, function=None, show=None, compare=None, one_liner_function=None,
                 name=None, is_root=False):
        """
        :param function: The function defining the experiment
        :param display_function: A function that can be called to display the results returned by function.
            This can be useful if you just want to re-run the display for already-computed and saved results.
            To do this, go experiment.save_last()
        :param conclusion: <Deprecated> will be removed in future
        :param name: Nmae of this experiment.
        """
        self.name = name
        self.function = function
        self._show = show
        self._one_liner_results = one_liner_function
        self._compare = compare
        self.variants = OrderedDict()
        self._notes = []
        self.is_root = is_root
        _register_experiment(self)

    @property
    def show(self):
        return self._show

    @property
    def one_liner_function(self):
        return self._one_liner_results

    @property
    def compare(self):
        return self._compare

    def __call__(self, *args, **kwargs):
        """ Run the function as normal, without recording or anything.  You can also modify with arguments. """
        return self.function(*args, **kwargs)

    def __str__(self):
        return 'Experiment {}'.format(self.name)

    def get_args(self):
        """
        :return: An OrderedDict of arguments to the experiment
        """
        return infer_derived_arg_values(self.function)

    def get_root_function(self):
        return get_partial_root(self.function)

    def is_generator(self):
        return inspect.isgeneratorfunction(self.get_root_function())

    def call(self, *args, **kwargs):
        """
        Call the experiment function without running as an experiment.  If the experiment is a function, this is the same
        as just result = my_exp_func().  If it's defined as a generator, it loops and returns the last result.
        :return: The last result
        """
        if self.is_generator():
            result = None
            for x in self(*args, **kwargs):
                result = x
        else:
            result = self(*args, **kwargs)
        return result

    def run(self, print_to_console=True, show_figs=None, test_mode=None, keep_record=None, raise_exceptions=True,
            display_results=False, notes = (), **experiment_record_kwargs):
        """
        Run the experiment, and return the ExperimentRecord that is generated.

        :param print_to_console: Print to console (as well as logging to file)
        :param show_figs: Show figures (as well as saving to file)
        :param test_mode: Run in "test_mode".  This sets the global "test_mode" flag when running the experiment.  This
            flag can be used to, for example, shorten a training session to verify that the code runs.  Can be:
            True: Run in test mode
            False: Don't run in test mode:
            None: Keep the current state of the global "is_test_mode()" flag.
        :param keep_record: Keep the folder that results are saved into.
            True: Results are saved into a folder
            False: Results folder is deleted at the end.
            None: If "test_mode" is true, then delete results at end, otherwise save them.
        :param raise_exceptions: True to raise any exception that occurs when running the experiment.  False to catch it,
            print the error, and move on.
        :param experiment_record_kwargs: Passed to the "record_experiment" context.
        :return: The ExperimentRecord object, if keep_record is true, otherwise None
        """
        if keep_record is None:
            keep_record = keep_record_by_default if keep_record_by_default is not None else not test_mode
        exp_rec = run_and_record(
            function = self.function,
            experiment_id=self.name,
            print_to_console=print_to_console,
            show_figs=show_figs,
            test_mode=test_mode,
            keep_record=keep_record,
            raise_exceptions=raise_exceptions,
            notes=notes,
            **experiment_record_kwargs
        )
        if display_results:
            self.show(exp_rec)

        return exp_rec

    def _create_experiment_variant(self, args, kwargs, is_root):
        assert len(args) in (0, 1), "When creating an experiment variant, you can either provide one unnamed argument (the experiment name), or zero, in which case the experiment is named after the named argumeents.  See add_variant docstring"
        name = args[0] if len(args) == 1 else _kwargs_to_experiment_name(kwargs)
        assert isinstance(name, str), 'Name should be a string.  Not: {}'.format(name)
        assert name not in self.variants, 'Variant "%s" already exists.' % (name,)
        ex = Experiment(
            name=self.name + '.' + name,
            function=partial(self.function, **kwargs),
            show=self._show,
            compare=self._compare,
            one_liner_function=self.one_liner_function,
            is_root=is_root
        )
        self.variants[name] = ex
        return ex

    def add_variant(self, variant_name = None, **kwargs):
        """
        Add a variant to this experiment, and register it on the list of experiments.
        There are two ways you can do this:

        .. code-block:: python

            # Name the experiment explicitely, then list the named arguments
            my_experiment_function.add_variant('big_a', a=10000)
            assert my_experiment_function.get_name()=='my_experiment_function.big_a'

            # Allow the experiment to be named automatically, and just list the named arguments
            my_experiment_function.add_variant(a=10000)
            assert my_experiment_function.get_name()=='my_experiment_function.a=10000'

        :param variant_name: Optionally, the name of the experiment
        :param kwargs: The named arguments which will differ from the base experiment.
        :return: The experiment.
        """
        return self._create_experiment_variant(() if variant_name is None else (variant_name, ), kwargs, is_root=False)

    def copy_variants(self, other_experiment):
        """
        Copy over the variants from another experiment.

        :param other_experiment: An Experiment Object
        """
        base_args = other_experiment.get_args()
        for variant in other_experiment.get_variants():
            if variant is not self:
                variant_args = variant.get_args()
                different_args = {k: v for k, v in variant_args.items() if base_args[k]!=v}
                name_diff = variant.get_id()[len(other_experiment.get_id())+1:]
                v = self.add_variant(name_diff, **different_args)
                v.copy_variants(variant)

    def add_root_variant(self, variant_name=None, **kwargs):
        """
        Add a variant to this experiment, but do NOT register it on the list of experiments.
        There are two ways you can do this:

        .. code-block:: python

            # Name the experiment explicitely, then list the named arguments
            my_experiment_function.add_root_variant('big_a', a=10000)
            assert my_experiment_function.get_name()=='my_experiment_function.big_a'

            # Allow the experiment to be named automatically, and just list the named arguments
            my_experiment_function.add_root_variant(a=10000)
            assert my_experiment_function.get_name()=='my_experiment_function.a=10000'

        :param variant_name: Optionally, the name of the experiment
        :param kwargs: The named arguments which will differ from the base experiment.
        :return: The experiment.
        """
        return self._create_experiment_variant(() if variant_name is None else (variant_name, ), kwargs, is_root=True)

    def get_id(self):
        """
        :return: A string uniquely identifying this experiment.
        """
        return self.name

    def get_variant(self, variant_name=None, **kwargs):
        """
        Get a variant on this experiment.

        :param variant_name: The name of the variant, if it has one
        :param kwargs: Otherwise, the named arguments which were used to define the variant.
        :return: An Experiment object
        """
        if variant_name is None:
            variant_name = _kwargs_to_experiment_name(kwargs)
        else:
            assert len(kwargs)==0, 'If you provide a variant name ({}), there is no need to specify the keyword arguments. ({})'.format(variant_name, kwargs)
        assert variant_name in self.variants, "No variant '{}' exists.  Existing variants: {}".format(variant_name, self.variants.keys())
        return self.variants[variant_name]

    def get_records(self, only_completed=False):
        """
        Get all records associated with this experiment.

        :param only_completed: Only include records that have run to completion.
        :return: A list of ExperimentRecord objects.
        """
        records = [load_experiment_record(rid) for rid in experiment_id_to_record_ids(self.name)]
        if only_completed:
            records = [record for record in records if record.get_status()==ExpStatusOptions.FINISHED]
        return records

    def browse(self, command=None, catch_errors = False, close_after = False, filterexp=None, filterrec = None,
            view_mode ='full', raise_display_errors=False, run_args=None, keep_record=True, truncate_result_to=100,
            cache_result_string = False, remove_prefix = None, display_format='nested', **kwargs):
        """
        Open up the UI, which allows you to run experiments and view their results.

        :param command: Optionally, a string command to pass directly to the UI.  (e.g. "run 1")
        :param catch_errors: Catch errors that arise while running experiments
        :param close_after: Close after issuing one command.
        :param filterexp: Filter the experiments with this selection (see help for how to use)
        :param filterrec: Filter the experiment records with this selection (see help for how to use)
        :param view_mode: How to view experiments {'full', 'results'} ('results' leads to a narrower display).
        :param raise_display_errors: Raise errors that arise when displaying the table (otherwise just indicate that display failed in table)
        :param run_args: A dict of named arguments to pass on to Experiment.run
        :param keep_record: Keep a record of the experiment after running.
        :param truncate_result_to: An integer, indicating the maximum length of the result string to display.
        :param cache_result_string: Cache the result string (useful when it takes a very long time to display the results
            when opening up the menu - often when results are long lists).
        :param remove_prefix: Remove the common prefix on the experiment ids in the display.
        :param display_format: How experements and their records are displayed: 'nested' or 'flat'.  'nested' might be
            better for narrow console outputs.
        """
        from artemis.experiments.ui import browse_experiments
        browse_experiments(command = command, root_experiment=self, catch_errors=catch_errors, close_after=close_after,
            filterexp=filterexp, filterrec=filterrec,
            view_mode=view_mode, raise_display_errors=raise_display_errors, run_args=run_args, keep_record=keep_record,
            truncate_result_to=truncate_result_to, cache_result_string=cache_result_string, remove_prefix=remove_prefix,
            display_format=display_format, **kwargs)

    # Above this line is the core api....
    # -----------------------------------
    # Below this line are a bunch of convenience functions.

    def has_record(self, completed=True, valid=True):
        """
        Return true if the experiment has a record, otherwise false.
        :param completed: Check that the record is completed.
        :param valid: Check that the record is valid (arguments match current experiment arguments)
        :return: True/False
        """
        records = self.get_records(only_completed=completed)
        if valid:
            records = [record for record in records if record.args_valid()]
        return len(records)>0

    def get_variants(self):
        return self.variants.values()

    def get_all_variants(self, include_roots=False, include_self=True):
        """
        Return a list of variants of this experiment
        :param include_roots: Include "root" experiments
        :param include_self: Include this experiment (unless include_roots is false and this this experiment is a root)
        :return: A list of experiments.
        """
        variants = []
        if include_self and (not self.is_root or include_roots):
            variants.append(self)
        for name, v in self.variants.items():
            variants += v.get_all_variants(include_roots=include_roots, include_self=True)
        return variants

    def test(self, **kwargs):
        self.run(test_mode=True, **kwargs)

    def get_latest_record(self, only_completed=False, err_if_none = True):
        """
        Return the ExperimentRecord from the latest run of this Experiment.

        :param only_completed: Only search among records of that have run to completion.
        :param err_if_none: If True, raise an error if no record exists.  Otherwise, just return None in this case.
        :return: An ExperimentRecord object
        """
        records = self.get_records(only_completed=only_completed)
        if len(records)==0:
            if err_if_none:
                raise Exception('No{} records for experiment "{}"'.format(' completed' if only_completed else '', self.name))
            else:
                return None
        else:
            return sorted(records, key=lambda x: x.get_id())[-1]

    def get_variant_records(self, only_completed=False, only_last=False, flat=False):
        """
        Get the collection of records associated with all variants of this Experiment.

        :param only_completed: Only search among records of that have run to completion.
        :param only_last: Just return the most recent record.
        :param flat: Just return a list of records
        :return: if not flat (default) An OrderedDict<experiment_id: ExperimentRecord>.
            otherwise, if flat: a list<ExperimentRecord>
        """
        variants = self.get_all_variants(include_self=True)

        if only_last:
            exp_record_dict = OrderedDict((ex.name, ex.get_latest_record(only_completed=only_completed, err_if_none=False)) for ex in variants)
            if flat:
                return [record for record in exp_record_dict.values() if record is not None]
            else:
                return exp_record_dict
        else:
            exp_record_dict = OrderedDict((ex.name, ex.get_records(only_completed=only_completed)) for ex in variants)
            if flat:
                return [record for records in exp_record_dict.values() for record in records]
            else:
                return exp_record_dict


_GLOBAL_EXPERIMENT_LIBRARY = OrderedDict()


class ExperimentNotFoundError(Exception):
    def __init__(self, experiment_id):
        Exception.__init__(self,'Experiment "{}" could not be loaded, either because it has not been imported, or its definition was removed.'.format(experiment_id))


def clear_all_experiments():
    _GLOBAL_EXPERIMENT_LIBRARY.clear()


@contextmanager
def capture_created_experiments():
    """
    A convenient way to cross-breed experiments.  If you define experiments in this block, you can capture them for
    later use (for instance by modifying them). e.g.:

    .. code-block:: python

        @experiment_function
        def add_two_numbers(a=1, b=2):
            return a+b

        with capture_created_experiments() as exps:
            add_two_numbers.add_variant(a=2)
            add_two_numbers.add_variant(a=3)

        for ex in exps:
            ex.add_variant(b=4)

    :rtype: Generator[:class:`Experiment`]
    """
    current_len = len(_GLOBAL_EXPERIMENT_LIBRARY)
    new_experiments = []
    yield new_experiments
    for ex in _GLOBAL_EXPERIMENT_LIBRARY.values()[current_len:]:
        new_experiments.append(ex)


def _register_experiment(experiment):
    assert experiment.name not in _GLOBAL_EXPERIMENT_LIBRARY, 'You have already registered an experiment named {} in {}'.format(experiment.name, inspect.getmodule(experiment.get_root_function()).__name__)
    _GLOBAL_EXPERIMENT_LIBRARY[experiment.name] = experiment


def get_nonroot_global_experiment_library():
    return OrderedDict((name, exp) for name, exp in _GLOBAL_EXPERIMENT_LIBRARY.items() if not exp.is_root)


def get_experiment_info(name):
    experiment = load_experiment(name)
    return str(experiment)


def load_experiment(experiment_id):
    try:
        return _GLOBAL_EXPERIMENT_LIBRARY[experiment_id]
    except KeyError:
        raise ExperimentNotFoundError(experiment_id)


def is_experiment_loadable(experiment_id):
    assert isinstance(experiment_id, string_types), 'Expected a string for experiment_id, not {}'.format(experiment_id)
    return experiment_id in _GLOBAL_EXPERIMENT_LIBRARY


def _kwargs_to_experiment_name(kwargs):
    return ','.join('{}={}'.format(argname, kwargs[argname]) for argname in sorted(kwargs.keys()))


@contextmanager
def hold_global_experiment_libary(new_lib = None):
    if new_lib is None:
        new_lib = OrderedDict()

    global _GLOBAL_EXPERIMENT_LIBRARY
    oldlib = _GLOBAL_EXPERIMENT_LIBRARY
    _GLOBAL_EXPERIMENT_LIBRARY = new_lib
    yield _GLOBAL_EXPERIMENT_LIBRARY
    _GLOBAL_EXPERIMENT_LIBRARY = oldlib


def get_global_experiment_library():
    return _GLOBAL_EXPERIMENT_LIBRARY


keep_record_by_default = None


@contextmanager
def experiment_testing_context(close_figures_at_end = True, new_experiment_lib = False):
    """
    Use this context when testing the experiment/experiment_record infrastructure.
    Should only really be used in test_experiment_record.py
    """
    ids = get_all_record_ids()
    global keep_record_by_default
    old_val = keep_record_by_default
    keep_record_by_default = True
    if new_experiment_lib:
        with hold_global_experiment_libary():
            yield
    else:
        yield
    keep_record_by_default = old_val

    if close_figures_at_end:
        from matplotlib import pyplot as plt
        plt.close('all')

    def clean_on_close():
        new_ids = set(get_all_record_ids()).difference(ids)
        clear_experiment_records(list(new_ids))

    atexit.register(clean_on_close)  # We register this on exit to avoid race conditions with system commands when we open figures externally




import time
_call_counts = {}


class CallStatus(object):

    def __init__(self, name, print_time = None):
        self.name = name
        self.start_time = time.time()
        self.call_count = 0
        self.print_time = print_time
        self.last_print_time = self.start_time

    def update(self):
        current_time = time.time()
        self.call_count += 1

        if self.print_time is None or (current_time - self.last_print_time) > self.print_time:
            print('Timer {}: {} iterations in {:.3}s ({:.3} iterations/second)'.format(self.name, self.call_count, current_time-self.start_time, self.call_count/(current_time-self.start_time)))
            self.last_print_time = current_time


def report_call_timing(name, print_time=2):
    """
    Print a periodic report of the number of times this function is called per second.

    This can be useful when trying to assess frequently a loop is running in your code.

    :param name: A "name" for this timer - a unique name which identifies what you're trying to time.
    :param print_time: How often (in seconds) you'd like to print an update.
    :return:
    """

    if name not in _call_counts:
        _call_counts[name] = CallStatus(name, print_time=print_time)

    _call_counts[name].update()



import numpy as np

from artemis.fileman.file_getter import get_file
from artemis.general.should_be_builtins import memoize
from artemis.ml.datasets.datasets import DataSet


__author__ = 'peter'


@memoize
def get_20_newsgroups_dataset(filter_most_common = 2000, numeric = False, shuffling_seed = 1234, bag_of_words = False, count_scaling = None):
    """
    The 20 newsgroups dataset.  In this dataset, you try to predict the topic of a forum from the words contained in
    posts in the forums.

    Words have been preprocessed to the "stemmed" version, as explained on the website:
    http://ana.cachopo.org/datasets-for-single-label-text-categorization

    :param filter_most_common: Can be:
        None: Don't filter out words
        int N: Filter out words that are not in the N most common workds
        (int N, int M): Filter out words that are not between the Nth and Mth most common words.
    :param numeric: Convert everything from words to numbers
    :param shuffling_seed: Random seed for shuffling (you want to shuffle, because everything's sorted by topic)
    :param bag_of_words: Return count vectors for each word
    :param count_scaling: If using bag_of_words, apply the transformation:
        vector = log(1+word_counts)
        To generate the input data (this scaling makes it more suitable for some types of classifiers).
    :return: A DataSet object
    """

    training_set_file = get_file(
        relative_name = 'data/20ng-train-stemmed.txt',
        url = 'http://ana.cachopo.org/datasets-for-single-label-text-categorization/20ng-train-stemmed.txt'
        )

    test_set_file = get_file(
        relative_name = 'data/20ng-test-stemmed.txt',
        url = 'http://ana.cachopo.org/datasets-for-single-label-text-categorization/20ng-test-stemmed.txt'
        )

    train_words, train_labels = _read_formatted_file(training_set_file)
    test_words, test_labels = _read_formatted_file(test_set_file)

    # Shuffle it up...
    rng = np.random.RandomState(shuffling_seed)
    train_words, train_labels =_shuffle((train_words, train_labels), rng)
    test_words, test_labels =_shuffle((test_words, test_labels), rng)

    # Filter out most-common-but-not-too-common-words
    all_train_words = np.concatenate(train_words)
    filtered_vocab, counts = _find_most_common(all_train_words, filter_most_common)
    train_words = _filter_lists_of_words(train_words, filtered_vocab)
    test_words = _filter_lists_of_words(test_words, filtered_vocab)

    if numeric or bag_of_words:
        train_ixs_list = _list_of_posts_to_list_of_ixs(train_words, filtered_vocab)
        test_ixs_list = _list_of_posts_to_list_of_ixs(test_words, filtered_vocab)
        label_vocab = {lab: i for i, lab in enumerate(np.unique(train_labels))}
        train_labels = _words_to_ints(train_labels, label_vocab)
        test_labels = _words_to_ints(test_labels, label_vocab)

        if bag_of_words:
            train_counts = _list_of_ixs_to_count_matrix(train_ixs_list, n_words=len(filtered_vocab))
            test_counts = _list_of_ixs_to_count_matrix(test_ixs_list, n_words=len(filtered_vocab))
            if count_scaling == 'log':
                train_counts = np.log(1+train_counts)
                test_counts = np.log(1+test_counts)
            return DataSet.from_xyxy(training_inputs = train_counts, training_targets = train_labels, test_inputs = test_counts, test_targets = test_labels)
        else:
            return DataSet.from_xyxy(training_inputs = train_ixs_list, training_targets = train_labels, test_inputs = test_ixs_list, test_targets = test_labels)
    else:
        return DataSet.from_xyxy(training_inputs = train_words, training_targets = train_labels, test_inputs = test_words, test_targets = test_labels)



def _read_formatted_file(file_relative_path):

    with open(get_file(file_relative_path)) as f:
        text = f.read()
    pairs = [line.split('\t') for line in text.split('\n')[:-1]]
    labels = [group for group, _ in pairs]
    words = [sentence.split(' ') for _, sentence in pairs]
    return words, labels


def _find_most_common(elements, n_most_common):

    unique_elements, counts = np.unique(elements, return_counts=True)
    # ixs = np.argpartition(-counts, kth = n_most_common)
    ixs = np.argsort(counts)[::-1]
    if isinstance(n_most_common, int):
        most_common_element_ixs = ixs[:n_most_common]
    else:
        assert isinstance(n_most_common, tuple) and len(n_most_common) == 2, 'eh?'
        start, stop = n_most_common
        most_common_element_ixs = ixs[start:stop]
    most_common_elements = unique_elements[most_common_element_ixs]
    return most_common_elements, counts[most_common_element_ixs]


def _filter_words(word_list, filter_set):
    if not isinstance(filter_set, set):
        filter_set = set(filter_set)
    return [w for w in word_list if w in filter_set]


def _filter_lists_of_words(lists_of_words, filter_set):
    return np.array([_filter_words(word_list, filter_set) for word_list in lists_of_words])


def _words_to_ints(word_list, lookup):
    return np.array([lookup[w] for w in word_list])


def _list_of_posts_to_list_of_ixs(list_of_posts, vocabulary):
    div_ixs = np.cumsum([len(post) for post in list_of_posts])[:-1]
    all_filtered_words = np.concatenate(list_of_posts)
    ixs = np.zeros(len(all_filtered_words), dtype = int)
    for i, w in enumerate(vocabulary):
        ixs[all_filtered_words==w] = i
    list_of_ixs = np.split(ixs, div_ixs)
    return np.array(list_of_ixs)


def _list_of_ixs_to_count_matrix(list_of_ixs, n_words):
    n_samples = len(list_of_ixs)
    counts = np.zeros((n_samples, n_words), dtype = int)
    for c, ixs in zip(counts, list_of_ixs):
        np.add.at(c, ixs, 1)
    return counts


def _shuffle(arrays, rng):
    n_samples = len(arrays[0])
    assert all(n_samples == len(arr) for arr in arrays)
    ixs = np.array(rng.permutation(n_samples))
    return tuple(np.array(arr)[ixs] for arr in arrays)


if __name__ == '__main__':

    data = get_20_newsgroups_dataset(numeric=False, filter_most_common = (100, 500), bag_of_words=False)
    for _, (inputs, ), (targets, ) in data.training_set.shorten(20).minibatch_iterator(minibatch_size = 1):
        print('%s: %s' % (targets, inputs))



from tempfile import mkstemp
import cProfile
import pstats
from artemis.general.display import surround_with_header
import os


def what_are_we_waiting_for(command, sort_by ='time', max_len = 20, print_here = True):
    """
    An easy way to show what is taking all the time when you run something.
    Taken from docs: https://docs.python.org/2/library/profile.html#module-cProfile

    :param command: A string python command
    :param sort_by: How to sort results. {'time', 'cumtime', 'calls', ...}.
        See https://docs.python.org/2/library/profile.html#pstats.Stats.sort_stats
    :param max_len: Maximum number of things to show in profile.
    :param print_here: Print the results here (instead of returning them).
    :return: A pstats.Stats object containing the profiling results.
    """
    _, filepath = mkstemp()
    try:
        cProfile.run(command, filepath)

    finally:
        p = pstats.Stats(filepath)
        os.remove(filepath)
        p.strip_dirs()
        p.sort_stats(sort_by)
        if print_here:
            print(surround_with_header('Profile for "{}"'.format(command), width=100, char='='))
            p.print_stats(max_len)
            print('='*100)
        return p



from six.moves import queue
import threading
import time
from collections import namedtuple
from artemis.remote.utils import recv_size, send_size


def _queue_get_all_no_wait(q, max_items_to_retreive):
    """
    Empties the queue, but takes maximally maxItemsToRetreive from the queue
    :param q:
    :param max_items_to_retreive:
    :return:
    """
    items = []
    for numOfItemsRetrieved in range(0, max_items_to_retreive):
        try:
            items.append(q.get_nowait())
        except queue.Empty:
            break
    return items

def handle_socket_accepts(sock, main_input_queue=None, return_queue=None, max_number=0):
    """
    This Accepts max_number of incoming communication requests to sock and starts the threads that manages the data-transfer between the server and the clients
    :param sock:
    :param main_input_queue:
    :param return_queue:
    :param max_number:
    :return:
    """
    return_lock = threading.Lock()
    for _ in range(max_number):
        connection, client_address = sock.accept()
        if main_input_queue:
            t0 = threading.Thread(target=handle_input_connection,args=(connection, client_address, main_input_queue))
            t0.setDaemon(True)
            t0.start()

        if return_queue:
            t1 = threading.Thread(target=handle_return_connection,args=(connection, client_address, return_queue, return_lock))
            t1.setDaemon(True)
            t1.start()


def handle_return_connection(connection, client_address, return_queue, return_lock):
    """
    For each client, there is a thread that continously checks for the confirmation that a plot from this client has been rendered.
    This thread takes hold of the return queue, dequeues max 10 objects and checks if there is a return message for the client that is goverend by this thread.
    All other return messages are put back into the queue. Then the lock on the queue is released so that other threads might serve their clients their respecitve messages.
    The return messges belonging to this client are then sent back.
    :param connection:
    :param client_address:
    :param return_queue:
    :param return_lock:
    :return:
    """
    while True:
        return_lock.acquire()
        if not return_queue: break
        try:
            return_objects = _queue_get_all_no_wait(return_queue, 10)
        except Exception:
            break

        if len(return_objects) > 0:
            owned_items = []
            for client, plot_id in return_objects:
                if client == client_address:
                    owned_items.append(plot_id)
                else:
                    return_queue.put((client,plot_id))
            return_lock.release()
            for plot_id in owned_items:
                message = plot_id
                send_size(sock=connection, data=message)
        else:
            return_lock.release()
            time.sleep(0.01)

ClientMessage = namedtuple('ClientMessage', ['dbplot_message', 'client_address'])
# dbplot_args is a DBPlotMessage object
# client_address: A string IP address


def handle_input_connection(connection, client_address, input_queue):
    """
    For each client, there is a thread that waits for incoming plots over the network. If a plot came in, this plot is then put into the main queue from which the server takes
    plots away.
    :param connection:
    :param client_address:
    :param input_queue:
    :return:
    """
    while True:
        recv_message = recv_size(connection)
        if not input_queue: break
        input_queue.put(ClientMessage(recv_message, client_address))
    connection.close()



from collections import OrderedDict

from artemis.experiments.experiment_record_view import show_record, compare_experiment_records
from artemis.experiments.experiments import Experiment
from artemis.general.display import sensible_str
from artemis.general.should_be_builtins import uniquify_duplicates, izip_equal


def experiment_function(f):
    """
    Use this decorator (@experiment_function) on a function that you want to run.  e.g.

    .. code-block:: python

        @experiment_function
        def demo_my_experiment(a=1, b=2, c=3):
            ...

    This turns your function demo_my_experiment into an experiment.  It can still be called as a normal function, but
    it now has can also be called with the methods of an Experiment object (eg. demo_my_experiment.run()).

    """
    return ExperimentFunction()(f)


def experiment_root(f):
    """
    Use this decorator on a function that you want to build variants off of:

    .. code-block:: python

        @experiment_root
        def demo_my_experiment(a, b=2, c=3):
            ...

    The root experiment is not runnable by itself, and will not appear in the list in the browse experiments UI, but
    you can call ``demo_my_experiment.add_variant(...)`` to create runnable variants.
    """
    return ExperimentFunction(is_root=True)(f)


class ExperimentFunction(object):
    """
    This is the most general decorator.  You can use this to add details on the experiment.
    """

    def __init__(self, show = show_record, compare = compare_experiment_records, display_function=None, comparison_function=None, one_liner_function=sensible_str, is_root=False):
        """
        :param show:  A function that is called when you "show" an experiment record in the UI.  It takes an experiment
            record as an argument.
        :param compare: A function that is called when you "compare" a set of experiment records in the UI.
        :param display_function: [Deprecated] A function that takes the results (whatever your experiment returns) and displays them.
        :param comparison_function: [Deprecated] A function that takes an OrderedDict<experiment_name, experiment_return_value>.
            You can optionally define this function to compare the results of different experiments.
            You can use call this via the UI with the compare_experiment_results command.
        :param one_liner_function: A function that takes your results and returns a 1 line string summarizing them.
        :param is_root: True to make this a root experiment - so that it is not listed to be run itself.
        """
        self.show = show
        self.compare = compare

        if display_function is not None:
            assert show is show_record, "You can't set both display function and show.  (display_function is deprecated)"
            show = lambda rec: display_function(rec.get_result())

        if comparison_function is not None:
            assert compare is compare_experiment_records, "You can't set both display function and show.  (display_function is deprecated)"

            def compare(records):
                record_experiment_ids_uniquified = uniquify_duplicates(rec.get_experiment_id() for rec in records)
                comparison_function(OrderedDict((unique_rid, rec.get_result()) for unique_rid, rec in izip_equal(record_experiment_ids_uniquified, records)))

        self.show = show
        self.compare = compare
        self.is_root = is_root
        self.one_liner_function = one_liner_function

    def __call__(self, f):
        f.is_base_experiment = True
        ex = Experiment(
            name=f.__name__,
            function=f,
            show=self.show,
            compare = self.compare,
            one_liner_function=self.one_liner_function,
            is_root=self.is_root
        )
        return ex


import numpy as np

from artemis.general.mymath import cosine_distance, softmax


def get_evaluation_function(name):
    return {
        'mse': mean_squared_error,
        'mean_cosine_distance': lambda a, b: cosine_distance(a, b, axis=1).mean(),
        'mean_squared_error': mean_squared_error,
        'mean_l1_error': mean_l1_error,
        'percent_argmax_correct': percent_argmax_correct,
        'percent_argmax_incorrect': percent_argmax_incorrect,
        'percent_correct': percent_correct,
        'softmax_categorical_xe': softmax_categorical_xe
        }[name]


def mean_l1_error(actual, target):
    return np.mean(np.sum(np.abs(actual-target), axis=-1), axis=-1)


def mean_squared_error(actual, target):
    return np.mean(np.sum((actual-target)**2, axis = -1), axis = -1)


def softmax_categorical_xe(actual, target):
    """
    :param actual: An (n_samples, n_dims) array identifying pre-logistic output
    :param target: An (n_samples, ) integer array identifying labels
    :return:
    """
    if target.ndim==1:
        assert target.dtype==int and np.max(target) < actual.shape[1]
    elif target.ndim==2:
        assert np.all(target.sum(axis=1)==1) and np.all(np.max(target, axis=1)==1)
        target = np.argmax(target, axis=1)
    else:
        raise Exception("Don't know how to interpret a {}-D target".format(target))

    return np.mean(softmax(actual, axis=1)[np.arange(actual.shape[0]), target], axis=0)


def fraction_correct(actual, target):
    return np.mean(actual == target)


def percent_correct(actual, target):
    return 100*fraction_correct(actual, target)


def percent_argmax_correct(actual, target):
    """
    :param actual: An (n_samples, n_dims) array
    :param target: An (n_samples, ) array of indices OR an (n_samples, n_dims) array
    :return:
    """
    actual = collapse_onehot_if_necessary(actual)
    target = collapse_onehot_if_necessary(target)
    return 100*fraction_correct(actual, target)


def percent_binary_incorrect(actual, target):
    return 100.-percent_binary_correct(actual, target)


def percent_binary_correct(actual, target):
    """
    :param actual:  A (n_samples, ) array of floats between 0 and 1
    :param target: A (n_samples, ) array of True/False
    :return: The percent of times the "actual" was closes to the correct.
    """
    assert len(actual) == len(target)
    if actual.ndim==2:
        assert actual.shape[1]==1
        actual = actual[:, 0]
    if target.ndim==2:
        assert target.shape[1]==1
        target = target[:, 0]
    assert target.ndim==1
    if actual.ndim>1:
        assert actual.shape == (len(target), 1)
        actual = actual[:, 0]
    if np.array_equal(np.unique(target), (0, 1)):
        assert np.all(actual)>=0 and np.all(actual)<=1
        assert np.all((target==0)|(target==1))
        return 100*np.mean(np.round(actual) == target)
    elif np.array_equal(np.unique(target), (-1, 1)):
        assert np.all((target==-1)|(target==1))
        return 100*np.mean((actual>0)*2-1 == target)
    else:
        raise Exception("Go away I'm tired.")


def percent_argmax_incorrect(actual, target):
    return 100 - percent_argmax_correct(actual, target)


def collapse_onehot_if_necessary(output_data):
    """
    Given an input that could either be in onehot encoding or not, return it in onehot encoding.

    :param output_data: Either an (n_samples, n_dims) array, or an (n_samples, ) array of labels.
    :return: An (n_samples, ) array.
    """

    # output_data = np.squeeze(output_data)

    if output_data.ndim == 2:
        return np.argmax(output_data, axis = 1)
    else:
        assert output_data.ndim == 1 and output_data.dtype in (int, 'int32', bool)
        return output_data


import sys
from artemis.fileman.local_dir import get_artemis_data_path
from artemis.general.display import CaptureStdOut

__author__ = 'peter'

"""
Save Print statements:

Useful in ipython notebooks where you lose output when printing to the browser.

On advice from:
http://stackoverflow.com/questions/4675728/redirect-stdout-to-a-file-in-python

** Note this is no longer being used.  Possibly delete
"""


_ORIGINAL_STDOUT = sys.stdout
_ORIGINAL_STDERR = sys.stderr


def capture_print(log_file_path = 'logs/dump/%T-log.txt', print_to_console=True):
    """
    :param log_file_path: Path of file to print to, if (state and to_file).  If path does not start with a "/", it will
        be relative to the data directory.  You can use placeholders such as %T, %R, ... in the path name (see format
        filename)
    :param print_to_console:
    :param print_to_console: Also continue printing to console.
    :return: The absolute path to the log file.
    """
    local_log_file_path = get_artemis_data_path(log_file_path)
    logger = CaptureStdOut(log_file_path=local_log_file_path, print_to_console=print_to_console)
    logger.__enter__()
    sys.stdout = logger
    sys.stderr = logger
    return local_log_file_path


def stop_capturing_print():
    sys.stdout = _ORIGINAL_STDOUT
    sys.stderr = _ORIGINAL_STDERR


def new_log_file(log_file_path = 'dump/%T-log', print_to_console = False):
    """
    Just capture-print with different defaults - intended to be called from notebooks where
    you don't want all output printed, but want to be able to see it with a link.
    :param log_file_path: Path to the log file - %T is replaced with time
    :param print_to_console: True to continue printing to console
    """
    return capture_print(log_file_path=log_file_path, print_to_console=print_to_console)


def read_print():
    return sys.stdout.read()


def reprint():
    assert isinstance(sys.stdout, CaptureStdOut), "Can't call reprint unless you've turned on capture_print"
    # Need to avoid exponentially growing prints...
    current_stdout = sys.stdout
    sys.stdout = _ORIGINAL_STDOUT
    print(read_print())
    sys.stdout = current_stdout



import numpy as np
from artemis.experiments import ExperimentFunction
from matplotlib import pyplot as plt
from six.moves import xrange


def display_drunkards_walk(drunkards):
    plt.plot(drunkards[:, :, 0], drunkards[:, :, 1])
    plt.grid()
    plt.xlabel('$\Delta$ Longitude (arcseconds)')
    plt.ylabel('$\Delta$ Latitude (arcseconds)')
    plt.show()


def compare_drunkards_walk(dict_of_drunkards):
    plot_handles = []
    for i, (exp_name, drunkards) in enumerate(dict_of_drunkards.items()):
        plot_handles.append(plt.plot(drunkards[:, :, 0], drunkards[:, :, 1], color='C{}'.format(i)))
    plt.grid()
    plt.xlabel('$\Delta$ Longitude (arcseconds)')
    plt.ylabel('$\Delta$ Latitude (arcseconds)')
    plt.legend([p[0] for p in plot_handles], dict_of_drunkards.keys())
    plt.show()


@ExperimentFunction(display_function=display_drunkards_walk, comparison_function=compare_drunkards_walk)
def demo_drunkards_walk(n_steps=500, n_drunkards=5, homing_instinct = 0, n_dim=2, seed=1234):
    """
    Release several drunkards in a field to randomly stumble around.  Record their progress.
    """
    rng = np.random.RandomState(seed)
    drunkards = np.zeros((n_steps+1, n_drunkards, n_dim))
    for t in xrange(1, n_steps+1):
        drunkards[t] = drunkards[t-1]*(1-homing_instinct) + rng.randn(n_drunkards, n_dim)
        if t%100==0:
            print('Status at step {}: Mean: {}, STD: {}'.format(t, drunkards[t].mean(), drunkards[t].std()))
    return drunkards


demo_drunkards_walk.add_variant(homing_instinct = 0.01)
demo_drunkards_walk.add_variant(homing_instinct = 0.1)


if __name__ == '__main__':
    demo_drunkards_walk.browse()
    # Try
    #   run all
    #   compare all
    #   display 1



from artemis.general.duck import Duck
from artemis.general.progress_indicator import ProgressIndicator
from artemis.general.checkpoint_counter import Checkpoints
import numpy as np
import time


def train_and_test_predictor(
        f_train,
        f_predict,
        losses,
        training_data_gen,
        test_data_gen_constructors,
        n_training_iters = None,
        n_test_iters = None,
        test_checkpoints = ('lin', 1000),
        collapse_loss = 'mean',
        progress_update_period = '5s',
        in_test_callback = None,
        post_test_callback = None,
        post_train_callback = None,
        save_train_return = False,
        measures = None,
        iterations_to_end = False
        ):
    """
    :param f_train:
    :param f_predict:
    :param losses:
    :param training_data_gen:
    :param test_data_gen_constructors:
    :param samples_count:
    :param n_training_iters:
    :param n_test_iters:
    :param test_checkpoints:
    :param collapse_loss:
    :param progress_update_period:
    :param in_test_callback:
    :param post_test_callback:
    :param post_train_callback:
    :param save_train_return:
    :param measures:
    :return: If yield_array_data is False, an ArrayStruct with fields:
        'training'                          Results recorded during training callbacks
            training_iter                   The iteration of the training callback
        'testing'                           Results recorded during tests
            test_iter                       The index of the test
                'iter'                      The number of training iterations finished at the time that the test is run
                'time'                      The time at which the test is run
                'samples'                   The number of samples seen so far
                'results'                   A structure containing results
                    subset_name             The name of the testing subset e.g. 'train', 'test' (you give subset names in test_data_gen_constructors)
                        'losses'
                            loss_name       The name of the loss function (you provide this in losses)
                            n_tests         The numner of tests that were run for this subset
                        'time'              The time, in seconds, that it took to test on this subset.


        Otherwise, if true, a structure the same object but training_iter and test_iter pushed to the leaf-position
    """

    if measures is None:
        measures = Duck()
    if 'training' not in measures:
        measures['training'] = Duck()
    if 'testing' not in measures:
        measures['testing'] = Duck()

    is_test_time = Checkpoints(test_checkpoints) if not isinstance(test_checkpoints, Checkpoints) else test_checkpoints
    pi = ProgressIndicator(n_training_iters, "Training", update_every=progress_update_period)

    for inputs, targets in training_data_gen:
        if is_test_time():

            this_test_measures = measures['testing'].open(next)
            this_test_measures['iter'] = pi.get_iterations()
            this_test_measures['time'] = pi.get_elapsed()
            this_test_measures['results'] = do_test(
                test_subset_generators={subset_name: constructor() for subset_name, constructor in test_data_gen_constructors.items()},
                f_predict=f_predict,
                loss_dict=losses,
                n_test_iters=n_test_iters,
                collapse_loss = collapse_loss,
                in_test_callback = in_test_callback,
                )
            if post_test_callback is not None:
                post_test_callback(this_test_measures)
            if iterations_to_end:
                measures_to_yield = measures.arrayify_axis(axis=1, subkeys='training')
                measures_to_yield = measures_to_yield.arrayify_axis(axis=1, subkeys='testing', inplace=True)
                yield measures_to_yield.to_struct()
            else:
                yield measures

        train_return = f_train(inputs, targets)
        pi.print_update()
        if save_train_return:
            measures['training', 'returns'] = train_return
        if post_train_callback:
            return_val = post_train_callback(inputs=inputs, targets=targets, iter=pi.get_iterations())
            if return_val is not None:
                measures['training', 'callback', next, ...] = return_val


def do_test(test_subset_generators, f_predict, loss_dict, n_test_iters, collapse_loss = 'mean', in_test_callback = None):

    if callable(loss_dict):
        loss_dict = dict(loss=loss_dict)

    these_test_results = Duck()
    losses = {}
    pi = ProgressIndicator(n_test_iters, "Testing")
    for subset_name, subset_generator in test_subset_generators.items():
        start_time = time.time()
        losses[subset_name] = []
        n_tests = 0
        for inputs, targets in subset_generator:
            n_tests+=1
            outputs = f_predict(inputs)
            for loss_name, f_loss in loss_dict.items():
                these_test_results[subset_name, 'losses', loss_name, next] = f_loss(outputs, targets)
            pi.print_update()
            if in_test_callback is not None:
                in_test_callback(inputs=inputs, targets=targets, outputs=outputs)

        assert n_tests>0, "It appears that subset '{}' had no tests!".format(subset_name)
        these_test_results[subset_name, 'n_tests'] = n_tests
        if collapse_loss is not None:
            collapse_func = {'mean': np.mean}[collapse_loss]
            for loss_name, f_loss in loss_dict.items():
                these_test_results[subset_name, 'losses', loss_name] = collapse_func(these_test_results[subset_name, 'losses', loss_name])
        these_test_results[subset_name, 'time'] = time.time() - start_time
    return these_test_results



from collections import OrderedDict
from contextlib import contextmanager

import itertools
import matplotlib.pyplot as plt
from artemis.general.should_be_builtins import bad_value
from artemis.plotting.data_conversion import vector_length_to_tile_dims

__author__ = 'peter'


_newplot_settings = dict(
    layout='grid',
    grid = False,
    show_x = True,
    show_y = True,
    xscale = None,
    yscale = None,
    sharex = None,
    sharey = None,
    xlabel = None,
    ylabel = None,
    xlim = None,
    ylim = None,
    )


def _create_subplot(fig = None, layout = None, position = None, **subplot_args):

    if layout is None:
        layout = _newplot_settings['layout']
    if fig is None:
        fig = plt.gcf()
    n = len(fig.axes)
    n_rows, n_cols = (1, n+1) if layout in ('h', 'horizontal') else (n+1, 1) if layout in ('v', 'vertical') else \
        vector_length_to_tile_dims(n+1) if layout in ('g', 'grid') else bad_value(layout)
    for i in range(n):
        fig.axes[i].change_geometry(n_rows, n_cols, i+1)

    for arg in ('sharex', 'sharey'):
        if isinstance(_newplot_settings[arg], plt.Axes):
            subplot_args[arg]=_newplot_settings[arg]

    ax = fig.add_subplot(n_rows, n_cols, n+1, **subplot_args)

    if _newplot_settings['xlabel'] is not None:
        ax.set_xlabel(_newplot_settings['xlabel'])
    if _newplot_settings['ylabel'] is not None:
        ax.set_ylabel(_newplot_settings['ylabel'])

    if _newplot_settings['xlim'] is not None:
        ax.set_xlim(_newplot_settings['xlim'])
    if _newplot_settings['ylim'] is not None:
        ax.set_ylim(_newplot_settings['ylim'])

    if _newplot_settings['grid']:
        plt.grid()

    for arg in ('sharex', 'sharey'):
        if _newplot_settings[arg] is True:
            _newplot_settings[arg]=ax

    if not _newplot_settings['show_x']:
        ax.tick_params(axis='x', labelbottom='off')
        # ax.get_xaxis().set_visible(False)
    if not _newplot_settings['show_y']:
        ax.tick_params(axis='y', labelleft='off')
        # ax.get_yaxis().set_visible(False)
    return ax



_subplots = OrderedDict()

_plot_name_generator = ('_autogenerated_subplot_{}'.format(i) for i in itertools.count(0))


def select_subplot(name, fig=None, layout=None, **subplot_args):
    """
    Set the current axes.  If "name" has been defined, just return that axes, otherwise make a new one.

    :param name: The name of the subplot
    :param fig: The figure, or None to select current figure
    :param layout: 'h' for horizontal layout, 'v' for vertical layout, 'g' for approximately-square grid
    :return: An axes object
    """

    if fig is None:
        fig = plt.gcf()

    if name in _subplots and fig is _subplots[name].get_figure():
        # (subplot has been created) and (figure containing it has not been closed)
        plt.subplot(_subplots[name])
    else:
        if name is None:
            name = next(_plot_name_generator)
        _subplots[name] = _create_subplot(fig=fig, layout=layout, **subplot_args)
    return _subplots[name]


def add_subplot(layout = None, fig = None, **subplot_args):
    """
    Add a subplot, and adjust the positions of the other subplots appropriately.
    Lifted from this answer: http://stackoverflow.com/a/29962074/851699

    :param fig: The figure, or None to select current figure
    :param layout: 'h' for horizontal layout, 'v' for vertical layout, 'g' for approximately-square grid
    :return: A new axes object
    """
    return select_subplot(name=None, fig=fig, layout=layout, **subplot_args)


@contextmanager
def _define_plot_settings(**settings):
    global _newplot_settings
    old_settings = _newplot_settings
    _newplot_settings = _newplot_settings.copy()
    _newplot_settings.update(settings)
    yield
    _newplot_settings = old_settings


class CaptureNewSubplots(object):

    def __enter__(self):
        self._old_subplots = _subplots.copy()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.new_subplots = OrderedDict((s, p) for s, p in _subplots.items() if s not in self._old_subplots)

    def get_new_subplots(self):
        return self.new_subplots


@contextmanager
def hstack_plots(spacing=0):

    with CaptureNewSubplots() as cap:
        with _define_plot_settings(layout='h', show_y = False):
            plt.subplots_adjust(wspace=spacing)
            yield
    new_subplots = cap.get_new_subplots().values()
    new_subplots[0].get_yaxis().set_visible(True)
    for ax in new_subplots[:-1]:
        ax.set_xticks(ax.get_xticks()[:-1])


def set_same_xlims(axes):
    xmins, xmaxs = zip(*[ax.get_xlim() for ax in axes])
    x_range = min(xmins), max(xmaxs)
    for ax in axes:
        ax.set_ylim(x_range)


def set_same_ylims(axes):
    ymins, ymaxs = zip(*[ax.get_ylim() for ax in axes])
    y_range = min(ymins), max(ymaxs)
    for ax in axes:
        ax.set_ylim(y_range)


def set_figure_border_size(size=0.05):
    plt.subplots_adjust(left=size, right=1.-size, top=1-size, bottom=size)

@contextmanager
def hstack_plots(spacing=0, sharex=False, sharey = True, grid=False, show_x=True, show_y='once', clip_x=False, clip_y=False, remove_ticks = True, xlabel=None, ylabel=None, xlim=None, ylim=None, **adjust_kwargs):

    with CaptureNewSubplots() as cap:
        with _define_plot_settings(layout='h', show_y = False if show_y=='once' else show_y, show_x = show_x, grid=grid, sharex=sharex, sharey=sharey, xlabel=xlabel, xlim=xlim, ylim=ylim):
            plt.subplots_adjust(wspace=spacing, **adjust_kwargs)
            yield
    new_subplots = cap.get_new_subplots().values()

    if clip_x:
        set_same_xlims(new_subplots)
    if clip_y:
        set_same_ylims(new_subplots)

    assert len(new_subplots)>0, "No new plots have been created in this block... Why did you create the block at all?"
    if show_y in (True, 'once'):
        new_subplots[0].tick_params(axis='y', labelleft='on')
    new_subplots[0].set_ylabel(ylabel)

    if remove_ticks:
        for ax in new_subplots[:-1]:
            ax.set_xticks(ax.get_xticks()[:-1])


@contextmanager
def vstack_plots(spacing=0, sharex=True, sharey = False, show_x = 'once', show_y=True, clip_x=False, clip_y=False, grid=False, remove_ticks = True, xlabel=None, ylabel=None, xlim=None, ylim=None, **adjust_kwargs):

    with CaptureNewSubplots() as cap:
        with _define_plot_settings(layout='v', show_x = False if show_x=='once' else show_x, show_y=show_y, grid=grid, sharex=sharex, sharey=sharey, ylabel=ylabel, xlim=xlim, ylim=ylim):
            plt.subplots_adjust(hspace=spacing, **adjust_kwargs)
            yield
    new_subplots = cap.get_new_subplots().values()

    if clip_x:
        set_same_xlims(new_subplots)
    if clip_y:
        set_same_ylims(new_subplots)

    assert len(new_subplots)>0, "No new plots have been created in this block... Why did you create the block at all?"
    if show_x in (True, 'once'):
        new_subplots[-1].tick_params(axis='x', labelbottom='on')
    new_subplots[-1].set_xlabel(xlabel)

    if remove_ticks:
        new_subplots[-1].get_xaxis().set_visible(True)
        for ax in new_subplots[:-1]:
            ax.set_yticks(ax.get_yticks()[:-1])



from collections import OrderedDict
from artemis.general.nested_structures import flatten_struct
from artemis.plotting.data_conversion import vector_length_to_tile_dims
import numpy as np
from matplotlib import pyplot as plt
from artemis.plotting.matplotlib_backend import get_plot_from_data

__author__ = 'peter'


def ezplot(anything, plots = None, hang = True, **plot_preference_kwargs):
    """
    Make a plot of anything.  Anything at all.
    :param anything: Anything.
    """
    data_dict = flatten_struct(anything)
    figure, plots = plot_data_dict(data_dict, plots, mode = 'static', hang = hang, **plot_preference_kwargs)
    return figure, plots


def plot_data_dict(data_dict, plots = None, mode = 'static', hang = True, figure = None, size = None, **plot_preference_kwargs):
    """
    Make a plot of data in the format defined in data_dict
    :param data_dict: dict<str: plottable_data>
    :param plots: Optionally, a dict of <key: IPlot> identifying the plot objects to use (keys should
        be the same as those in data_dict).
    :return: The plots (same ones you provided if you provided them)
    """

    assert mode in ('live', 'static')
    if isinstance(data_dict, list):
        assert all(len(d) == 2 for d in data_dict), "You can provide data as a list of 2 tuples of (plot_name, plot_data)"
        data_dict = OrderedDict(data_dict)

    if plots is None:
        plots = {k: get_plot_from_data(v, mode = mode, **plot_preference_kwargs) for k, v in data_dict.items()}

    if figure is None:
        if size is not None:
            from pylab import rcParams
            rcParams['figure.figsize'] = size
        figure = plt.figure()
    n_rows, n_cols = vector_length_to_tile_dims(len(data_dict))
    for i, (k, v) in enumerate(data_dict.items()):
        plt.subplot(n_rows, n_cols, i + 1)
        plots[k].update(v)
        plots[k].plot()
        plt.title(k, fontdict = {'fontsize': 8})
    oldhang = plt.isinteractive()
    plt.interactive(not hang)
    plt.show()
    plt.interactive(oldhang)
    return figure, plots


def funplot(func, xlims = None, n_points = 100, keep_ylims = False, **plot_args):
    """
    Plot a function
    :param func:
    :param xlims:
    :param n_points:
    :return:
    """
    if xlims is None:
        xlims = plt.gca().get_xbound()
    xs, xe = xlims
    x = np.linspace(xs, xe, n_points)
    if keep_ylims:
        ylims = plt.gca().get_ybound()
    plt.plot(x, func(x), **plot_args)
    if keep_ylims:
        plt.gca().set_ybound(*ylims)
    plt.gca().set_xbound(*xlims)



from artemis.general.hashing import compute_fixed_hash, fixed_hash_eq
import numpy as np
import sys

_IS_PYTHON_3 = sys.version_info > (3, 0)


def test_compute_fixed_hash():
    # Not really sure why the fixed hash differes between python 2 and 4 here (maybe something do do with changes to strings)

    complex_obj = [1, 'd', {'a': 4, 'b': np.arange(10)}, (7, list(range(10)))]
    original_code = compute_fixed_hash(complex_obj)

    expected_code = 'c9b83dd2e1099c3bcbb05e3c69327c72' if _IS_PYTHON_3 else '6c98fabc301361863f321f6149a8a12a'

    assert compute_fixed_hash(complex_obj) == original_code == expected_code
    complex_obj[2]['b'][6]=0

    expected_code = 'a783b1f098fca8ccaf977a3123be5ac4' if _IS_PYTHON_3 else '8aee1f739fc9a612ed72e14682026627'
    assert compute_fixed_hash(complex_obj) == expected_code != original_code
    complex_obj[2]['b'][6]=6  # Revert to old value
    assert compute_fixed_hash(complex_obj) == original_code


def test_compute_fixed_hash_terminates():

    a = []
    b = [a]
    a.append(b)
    code = compute_fixed_hash(a)
    assert code == 'cffaee424a62cd1893825a5811c34b8d'

    c = []
    d = [c]
    c.append(d)
    code = compute_fixed_hash(c)
    assert code == 'cffaee424a62cd1893825a5811c34b8d'


def test_fixed_hash_eq():

    obj1 = [1, 'd', {'a': 4, 'b': np.arange(10)}, (7, [1, 2, 3, 4, 5])]
    obj2 = [1, 'd', {'a': 4, 'b': np.arange(10)}, (7, [1, 2, 3, 4, 5])]
    obj3 = [1, 'd', {'a': 4, 'b': np.arange(10)}, (7, [1, 2, 3, 4, 5])]
    obj3[2]['b'][4] = 0
    assert fixed_hash_eq(obj1, obj2)
    assert not fixed_hash_eq(obj1, obj3)


if __name__ == '__main__':
    test_compute_fixed_hash()
    test_compute_fixed_hash_terminates()
    test_fixed_hash_eq()


from artemis.general.should_be_builtins import memoize
import numpy as np
from six.moves import xrange


__author__ = 'peter'


def vector_length_to_tile_dims(vector_length, ):
    """
    You have vector_length tiles to put in a 2-D grid.  Find the size
    of the grid that best matches the desired aspect ratio.

    TODO: Actually do this with aspect ratio

    :param vector_length:
    :return: n_rows, n_cols
    """
    n_cols = np.ceil(np.sqrt(vector_length))
    n_rows = np.ceil(vector_length/n_cols)
    grid_shape = int(n_rows), int(n_cols)
    return grid_shape


def put_vector_in_grid(vec, shape = None):
    if shape is None:
        n_rows, n_cols = vector_length_to_tile_dims(len(vec))
    else:
        n_rows, n_cols = shape
    grid = np.zeros(n_rows*n_cols, dtype = vec.dtype)
    grid[:len(vec)]=vec
    grid=grid.reshape(n_rows, n_cols)
    return grid


@memoize
def _data_shape_and_boundary_width_to_grid_slices(shape, grid_shape, boundary_width, is_colour = None):

    assert len(shape) in (3, 4) or len(shape)==5 and shape[-1]==3
    if is_colour is None:
        is_colour = shape[-1]==3
    size_y, size_x = (shape[-3], shape[-2]) if is_colour else (shape[-2], shape[-1])
    is_vector = (len(shape)==4 and is_colour) or (len(shape)==3 and not is_colour)

    if grid_shape is None:
        grid_shape = vector_length_to_tile_dims(shape[0]) if is_vector else shape[:2]
    n_rows, n_cols = grid_shape

    output_shape = n_rows*(size_y+boundary_width)+1, n_cols*(size_x+boundary_width)+1
    index_pairs = []
    for i in xrange(n_rows):
        for j in xrange(n_cols):
            if is_vector:
                pull_indices = (i*n_cols + j, )
                if pull_indices[0] == shape[0]:
                    break
            else:
                pull_indices = (i, j)
            start_row, start_col = i*(size_y+1)+1, j*(size_x+1)+1
            push_indices = slice(start_row, start_row+size_y), slice(start_col, start_col+size_x)
            index_pairs.append((pull_indices, push_indices))
    return output_shape, index_pairs


def put_data_in_grid(data, grid_shape = None, fill_colour = np.array((0, 0, 128), dtype = 'uint8'), cmap = 'gray',
        boundary_width = 1, clims = None, is_color_data=None, nan_colour=None):
    """
    Given a 3-d or 4-D array, put it in a 2-d grid.
    :param data: A 4-D array of any data type
    :return: A 3-D uint8 array of shape (n_rows, n_cols, 3)
    """
    output_shape, slice_pairs = _data_shape_and_boundary_width_to_grid_slices(data.shape, grid_shape, boundary_width, is_colour=is_color_data)
    output_data = np.empty(output_shape+(3, ), dtype='uint8')
    output_data[..., :] = fill_colour  # Maybe more efficient just to set the spaces.
    scaled_data = data_to_image(data, clims = clims, cmap = cmap, is_color_data=is_color_data, nan_colour=nan_colour)
    for pull_slice, push_slice in slice_pairs:
        output_data[push_slice] = scaled_data[pull_slice]
    return output_data


def put_list_of_images_in_array(list_of_images, fill_colour = np.array((0, 0, 0))):
    """
    Arrange a list of images into a grid.  They do not necessairlily need to have the same size.

    :param list_of_images: A list of images, not necessarily the same size.
    :param fill_colour: The colour with which to fill the gaps
    :return: A (n_images, size_y, size_x, 3) array of images.
    """
    size_y = max(im.shape[0] for im in list_of_images)
    size_x = max(im.shape[1] for im in list_of_images)
    im_array = np.zeros((len(list_of_images), size_y, size_x, 3))+fill_colour
    for g, im in zip(im_array, list_of_images):
        top = int((size_y-im.shape[0])/2)
        left = int((size_x-im.shape[1])/2)
        g[top:top+im.shape[0], left:left+im.shape[1], :] = im if im.ndim==3 else im[:, :, None]
    return im_array


def put_list_of_lists_of_images_in_array(list_of_lists_of_images, fill_colour = np.array((0, 0, 0))):
    """
    Arrange a list of lists of images into a grid.  Each sublist does not necessarily need to have the same length, and
    images within the lists do not necessairlily need to have the same size.

    :param list_of_lists_of_images: A list of lists of images.
    :param fill_colour: The colour with which to fill the gaps
    :return: A (n_rows, n_cols, size_y, size_x, 3) array of images.
    """
    image_arrays = [put_list_of_images_in_array(list_of_images) for list_of_images in list_of_lists_of_images]
    image_grid_tensor = np.zeros((len(image_arrays), max(arr.shape[0] for arr in image_arrays), max(arr.shape[1] for arr in image_arrays), max(arr.shape[2] for arr in image_arrays), 3))+fill_colour  # (n_rows, n_cols, size_y, size_x, 3)
    for i, arr in enumerate(image_arrays):
        image_grid_tensor[i, :arr.shape[0], :arr.shape[1], :arr.shape[2], :] = arr
    return image_grid_tensor


def scale_data_to_8_bit(data, in_range = None):
    """
    Scale data to range [0, 255] and put in uint8 format.
    """
    return scale_data_to_range(data, in_range=in_range, out_range=(0, 255)).astype(np.uint8)


def scale_data_to_range(data, in_range = None, out_range = (0, 1), clip_to_range = True):
    """
    Scale your data into a new range.

    :param data: An ndarray
    :param in_range: The range that you'd like to scale your data from (if None, it's calculated automatically)
    :param out_range: The range you'd like to scale your data to.
    :param clip_to_range: True if you want to cut data outside of your range.
    :return: An ndarray with values in [out_range[0], out_range[1]]
    """
    out_scale = float(out_range[1]-out_range[0])
    out_shift = out_range[0]

    compute_scale = in_range is None
    if compute_scale:
        in_range = (np.nanmin(data), np.nanmax(data)) if data.size != 0 else (0, 1)
    smin, smax = in_range
    if smin==smax:
        smax += 1.
    scale = out_scale/(smax-smin)

    if np.isnan(scale):  # Data is all nans, or min==max
        return np.zeros_like(data)
    else:
        out = (data-smin)*scale
        if out_shift != 0:
            out += out_shift

        if not compute_scale and clip_to_range:
            out[out<out_range[0]] = out_range[0]
            out[out>out_range[1]] = out_range[1]
        return out


mappables = {}


def data_to_image(data, is_color_data = None, clims = None, cmap = 'gray', nan_colour=None):
    import matplotlib.cm as cm
    from matplotlib.colors import Normalize
    """
    Convert and ndarray of data into RGB pixel data.

    :param data: An ndarray of data.
    :param is_color_data: A boolean indicating whether this is colour data already.  If not specified we guess.
    :param clims: The range of values that the colour scale should cover.  Values outside this range will be
        clipped to fall in the range.  If None, calculate range from the data.
    :param cmap: Colormap - Use any of the names in matplotlib - eg ('gray', 'jet', 'Paired', 'cubehelix')
    :return: An ndarray of unt8 colour data.  Shape is: data.shape if is_color_data else data.shape+(3, )
    """

    if is_color_data is None:
        is_color_data = data.shape[-1] == 3
    if is_color_data:
        assert data.shape[-1] == 3, 'If data is specified as being colour data, the final axis must have length 3.'

    if not is_color_data:
        # Need to apply the cmap.
        if cmap == 'gray':
            # For speed, we handle this separately
            scaled_data = scale_data_to_8_bit(data, in_range=clims)
            scaled_data = np.concatenate([scaled_data[..., None]]*3, axis = scaled_data.ndim)
        else:
            if (clims, cmap) not in mappables:
                mappables[clims, cmap] = cm.ScalarMappable(cmap = cmap, norm = None if clims is None else Normalize(vmin=clims[0], vmax=clims[1]))
            cmap = mappables[clims, cmap]
            old_dim = data.shape
            if len(old_dim)>2:
                data = data.reshape((data.shape[0], -1))
            rgba = cmap.to_rgba(data)
            if len(old_dim)>2:
                rgba = rgba.reshape(old_dim+(4, ))
            scaled_data = (rgba[..., :-1]*255)
    else:
        scaled_data = scale_data_to_8_bit(data, in_range=clims).astype(np.uint8)

    if nan_colour is not None:
        scaled_data = np.where(np.any(np.isnan(data if is_color_data else data[..., None]), axis=-1)[..., None], np.array(nan_colour, dtype=np.uint8), scaled_data)

    return scaled_data


class RecordBuffer(object):

    def __init__(self, buffer_len, initial_value = np.NaN):
        self._buffer_len = buffer_len
        self._buffer = None
        self._ix = 0
        self._base_indices = np.arange(buffer_len)
        self._initial_value = initial_value

    def __call__(self, data):
        if self._buffer is None:
            shape = () if np.isscalar(data) else data.shape
            dtype_data = data+self._initial_value
            dtype = dtype_data.dtype if isinstance(dtype_data, np.ndarray) else type(dtype_data) if isinstance(dtype_data, (int, float, bool)) else object
            self._buffer = np.empty((self._buffer_len, )+shape, dtype = dtype)
            self._buffer[:] = self._initial_value
        self._buffer[self._ix] = data
        self._ix = (self._ix+1) % self._buffer_len
        return self._buffer[(self._base_indices+self._ix) % self._buffer_len]


class UnlimitedRecordBuffer(object):

    def __init__(self, expansion_factor = 2, initial_size=64):
        self._buffer = []
        self._index = 0
        self.initial_size = initial_size
        self.expansion_factor = expansion_factor

    def __call__(self, data):
        if self._index==len(self._buffer):
            data_shape = data.shape if isinstance(data, np.ndarray) else ()
            if len(self._buffer)==0:
                self._buffer = np.empty((self.initial_size, )+data_shape)
            else:
                new_buffer = np.empty((int(len(self._buffer)*self.expansion_factor), )+data_shape)
                new_buffer[:self._index] = self._buffer
                self._buffer = new_buffer
        self._buffer[self._index] = data
        self._index += 1
        return self._buffer[:self._index]




import time
from six.moves import _thread as thread
from six.moves import SimpleHTTPServer, socketserver
from artemis.fileman.local_dir import get_artemis_data_path
from artemis.plotting.manage_plotting import set_show_callback, set_draw_callback
import os
from matplotlib import pyplot as plt
import logging
import atexit
import shutil
import uuid

logging.basicConfig()
ARTEMIS_LOGGER = logging.getLogger('artemis')

__author__ = 'peter'


def _make_plot_html(plot_directory, update_period=1.):
    """
    Generate the html file which will show the plots.
    :param update_period: The period of the update, in seconds.
    :return:
    """

    # Todo: This seems very dependent on where/how artemis is installed
    with open(os.path.join(os.path.dirname(__file__), 'artemis_plots.html')) as f:
        html_template = f.read()

    raw_html = html_template.replace('{update_period}', '%s' % update_period)

    # dest = get_artemis_data_path(os.path.join(plot_directory, 'index.html'), make_local_dir=True)
    dest = os.path.join(plot_directory, 'index.html')
    with open(dest, 'w') as f:
        f.write(raw_html)
    return dest


class QuietHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):

    def log_request(self, code='-', size='-'):
        pass  # Overrides: # self.log_message('"%s" %s %s', self.requestline, str(code), str(size))


def _launch_on_first_available_port(first_port):
    Handler = QuietHTTPRequestHandler
    port = first_port
    while True:
        try:
            httpd = socketserver.TCPServer(('', port), Handler)
            # print 'Serving on port', port
            ARTEMIS_LOGGER.warn("Serving Plots at http://localhost:%s" % (port, ))
            httpd.serve_forever()
        except socketserver.socket.error as exc:
            if exc.args[0] == 48 or exc.args[0] == 98:
                ARTEMIS_LOGGER.info('Port', port, 'already in use')
                port += 1
            else:
                raise
        else:
            break


def _start_plotting_server(plot_directory, port = 8000, update_period=1.):
    dest = _make_plot_html(plot_directory, update_period=update_period)
    plot_directory, _ = os.path.split(dest)
    os.chdir(plot_directory)
    thread.start_new(_launch_on_first_available_port, (port, ))


class TimedFigureSaver(object):

    def __init__(self, fig_loc, update_period = 1):
        self.fig_loc = fig_loc
        self._last_update = -float('inf')
        self.update_period = update_period

    def __call__(self, fig=None, block=False):
        current_time = time.time()
        if current_time - self._last_update > self.update_period:
            if fig is None:
                plt.savefig(self.fig_loc)
            else:
                fig.savefig(self.fig_loc)
            self._last_update = current_time


def setup_web_plotting(update_period = 1.):
    plot_directory = get_artemis_data_path(relative_path="tmp/web_backend/%s/" % (str(uuid.uuid4()),), make_local_dir=True)  # Temporary directory
    atexit.register(clean_up,plot_dir=plot_directory)
    _start_plotting_server(plot_directory=plot_directory, update_period=update_period)
    set_draw_callback(TimedFigureSaver(os.path.join(plot_directory, 'artemis_figure.png'), update_period=update_period))
    set_show_callback(TimedFigureSaver(os.path.join(plot_directory, 'artemis_figure.png'), update_period=update_period))


def clean_up(plot_dir):
    shutil.rmtree(plot_dir)


from abc import abstractmethod
import numpy as np
from artemis.general.mymath import recent_moving_average
from six.moves import xrange

__author__ = 'peter'


class OneHotEncoding(object):

    def __init__(self, n_classes = None, form = 'bin', dtype = None):
        assert form in ('bin', 'sign')
        if dtype is None:
            dtype = np.int32 if form == 'sign' else bool
        self._n_classes = n_classes
        self._dtype = dtype
        self.form = form

    def __call__(self, data):
        if self._n_classes is None:
            self._n_classes = np.max(data)+1
        out = np.zeros((data.size, self._n_classes, ), dtype = self._dtype)
        if self.form == 'sign':
            out[:] = -1
        if data.size > 0:  # Silly numpy
            out[np.arange(data.size), data.flatten()] = 1
        out = out.reshape(data.shape+(self._n_classes, ))
        return out

    def inverse(self, data):
        return np.argmax(data, axis = 1)


class RunningAverage(object):

    def __init__(self):
        self._n_samples_seen = 0
        self._average = 0

    def __call__(self, data):
        self._n_samples_seen+=1
        frac = 1./self._n_samples_seen
        self._average = (1-frac)*self._average + frac*data
        return self._average

    @classmethod
    def batch(cls, x):
        return np.cumsum(x, axis=0)/np.arange(1, len(x)+1).astype(np.float)[(slice(None), )+(None, )*(x.ndim-1)]


class RecentRunningAverage(object):

    def __init__(self):
        self._n_samples_seen = 0
        self._average = 0

    def __call__(self, data):
        self._n_samples_seen+=1
        frac = 1/self._n_samples_seen**.5
        self._average = (1-frac)*self._average + frac*data
        return self._average

    @classmethod
    def batch(cls, x):
        # return recent_moving_average(x, axis=0)  # Works only for python 2.X, with weave
        ra = cls()
        return np.array([ra(x_) for x_ in x])


class RunningAverageWithBurnin(object):

    def __init__(self, burn_in_steps):
        self._burn_in_step_remaining = burn_in_steps
        self.averager = RunningAverage()

    def __call__(self, x):

        if self._burn_in_step_remaining > 0:
            self._burn_in_step_remaining-=1
            return x
        else:
            return self.averager(x)


class IDifferentiableFunction(object):

    @abstractmethod
    def __call__(self, *args):
        pass

    @abstractmethod
    def backprop_delta(self, delta_y):
        pass

    def batch_call(self, *args, **kwargs):
        return single_to_batch(self, *args, **kwargs)

    def batch_backprop_delta(self, *args, **kwargs):
        return single_to_batch(self.backprop_delta, *args, **kwargs)


class NonNormalize(IDifferentiableFunction):

    def __call__(self, x):
        return x

    def backprop_delta(self, delta_y):
        return delta_y


class RunningCenter(IDifferentiableFunction):
    """
    Keep an exponentially decaying running mean, subtract this from the value.
    """
    def __init__(self, half_life):
        self.decay_constant = np.exp(-np.log(2)/half_life)
        self.one_minus_decay_constant = 1-self.decay_constant
        self.running_mean = None

    def __call__(self, x):
        if self.running_mean is None:
            self.running_mean = np.zeros_like(x)
        self.running_mean[:] = self.decay_constant * self.running_mean + self.one_minus_decay_constant * x
        return x - self.running_mean

    def backprop_delta(self, delta_y):
        return self.decay_constant * delta_y



class RunningNormalize(IDifferentiableFunction):

    def __init__(self, half_life, eps = 1e-7, initial_std=1):
        self.decay_constant = np.exp(-np.log(2)/half_life)
        self.one_minus_decay_constant = 1-self.decay_constant
        self.running_mean = None
        self.eps = eps
        self.initial_std = initial_std

    def __call__(self, x):
        if self.running_mean is None:
            self.running_mean = np.zeros_like(x)
            self.running_mean_sq = np.zeros_like(x) + self.initial_std**2
        self.running_mean[:] = self.decay_constant * self.running_mean + self.one_minus_decay_constant * x
        self.running_mean_sq[:] = self.decay_constant * self.running_mean_sq + self.one_minus_decay_constant * x**2
        std = np.sqrt(self.running_mean_sq - self.running_mean**2)
        return (x - self.running_mean) / (std+self.eps)

    def backprop_delta(self, delta_y):
        """
        Ok, we're not doing this right at all, but lets just ignore the contribution of the current
        sample to the mean/std.  This makes the gradient waaaaaay simpler.  If you want to see the real thing, put

        (x-(a*u+(1-a)*x))/sqrt((a*s+(1-a)*x^2 - (a*u+(1-a)*x)^2))
        into http://www.derivative-calculator.net/
        (a stands for lambda here)

        :param delta_y: The derivative of the cost wrt the output of this normalizer
        :return: delta_x: The derivative of the cost wrt the input of this normalizer
        """
        std = np.sqrt(self.running_mean_sq - self.running_mean**2)
        return delta_y/std


def single_to_batch(fcn, *batch_inputs, **batch_kwargs):
    """
    :param fcn: A function
    :param batch_inputs: A collection of batch-form (n_samples, input_dims_i) inputs
    :return: batch_outputs, an (n_samples, output_dims) array
    """
    n_samples = len(batch_inputs[0])
    assert all(len(b) == n_samples for b in batch_inputs)
    first_out = fcn(*[b[0] for b in batch_inputs], **{k: b[0] for k, b in batch_kwargs.items()})
    if n_samples==1:
        return first_out[None]
    out = np.empty((n_samples, )+first_out.shape)
    out[0] = n_samples
    for i in xrange(1, n_samples):
        out[i] = fcn(*[b[i] for b in batch_inputs], **{k: b[i] for k, b in batch_kwargs.items()})
    return out



from __future__ import print_function
from six.moves import socketserver
import getpass
import logging
import socket
import struct
import sys

import os
from six.moves import input

from artemis.config import get_artemis_config_value

ARTEMIS_LOGGER = logging.getLogger('artemis')

def one_time_send_to(address, port, message):
    '''
    Send message to given address at port. This method is not intended to be used repeatedly at the same port as it does not maintain a socket connection

    :param address: A string IP address to send to.
    :param port: An integer port number, or a string which can be mapped to an integeer
    :param message: A string to send
    '''

    server_address = (address, int(port))
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect(tuple(server_address))
    except:
        raise

    send_size(sock, message)


def get_socket(address, port):
    '''
    Returns a socket, bound to the next free port starting from the given port.
    :param port:
    :return: Tuple (socket, used_port)
    '''
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    while True:
        try:
            server_address = (address, port)
            sock.bind(server_address)
        except socketserver.socket.error as exc:
            # ARTEMIS_LOGGER.info('Port', port, 'already in use')
            port += 1
            if exc.args[0] == 48 or exc.args[0] == 98:
                port += 1
            else:
                raise
        else:
            break
    return (sock,port)


def send_size(sock, data):
    try:
        sock.sendall(struct.pack('!I', len(data)))
        sock.sendall(data)
    except socketserver.socket.error as exc:
        if exc.args[0] == 32:
            print("Broken pipe", file=sys.stderr)
            sys.exit(0)
        else:
            raise


def recv_bytes(sock, size):
    buf = b''
    while size:
        try:
            newbuf = sock.recv(size)
        except socketserver.socket.error as exc:
            if exc.args[0] == 54:
                print("Connection reset by peer", file=sys.stderr)
                sys.exit(0)
            else:
                raise
        buf += newbuf
        size -= len(newbuf)
    return buf

def recv_size(sock):

    size_data = recv_bytes(sock, 4)
    size = struct.unpack('!I', size_data)[0]
    message = recv_bytes(sock,size)

    return message

def get_local_ips():
    '''
    Returns all the different ways that one can address the own machine.
    source: http://stackoverflow.com/a/274644/2068168
    :return: list of strings
    '''
    from netifaces import interfaces, ifaddresses, AF_INET

    ip_list = []
    for interface in interfaces():
        if AF_INET in ifaddresses(interface):
            for link in ifaddresses(interface)[AF_INET]:
                ip_list.append(link['addr'])
    return ip_list

def is_valid_ip(ip_address):
    '''
    checks if ip_address has a valid format
    :param address:
    :return:
    '''
    try:
        socket.inet_aton(ip_address)
        return True
    except:
        return False

def is_valid_port(port):
    '''
    checks if port is a valid string.
    This will make sure that any given string is numeric and between the range of 0 and 65535
    Source: http://stackoverflow.com/a/12968117/2068168
    :param port:
    :return:
    '''
    import re
    port_pattern = re.compile("^([0-9]{1,4}|[1-5][0-9]{4}|6[0-4][0-9]{3}|65[0-4][0-9]{2}|655[0-2][0-9]|6553[0-5])$")
    return port_pattern.match(port)


def get_remote_artemis_path(remote_ip):
    return get_artemis_config_value(
        section=remote_ip,
        option='artemis_path',
        default_generator=lambda: input('Specify Remote Artemis Installation Path: ').strip(),
        write_default=True
        )


def check_if_port_is_free(ip_address, port):
    '''
    This checks if the remote server is able to accept requests at the given port.
    :return:
    '''

    try:
        port = int(port)
    except ValueError:
        print ("Please provide a valid port for address %s. Received %s instead" %(ip_address, port))
        raise

    if ip_address in get_local_ips():
        import socket
        s=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.bind((ip_address,port))
        except:
            raise
        finally:
            s.close()
    else:
        check_ssh_connection(ip_address)
        ssh_connect = get_ssh_connection(ip_address=ip_address)
        check_port_function = 'python -c "import socket; s=socket.socket(socket.AF_INET, socket.SOCK_STREAM);s.bind((\'%s\',%i));s.close()"'%(ip_address,port)
        stdin , stdout, stderr = ssh_connect.exec_command(check_port_function)
        err = stderr.read()
        assert not err, "The remote address %s cannot allocate port %i. The following error was raised: \n %s" % (ip_address, port,err.strip().split("\n")[-1])


def get_ssh_connection(ip_address):
    '''
    This returns a ssh_connection to the given ip_address. Make sure to close the connection afterwards.
    Requires a public/private key to be set up with the remote system. The location of the private key can be
    specified in .artemisrc or, if not specified, will be looked for in ~/.ssh/id_rsa
    :param ip_address:
    :return:
    '''
    import paramiko
    path_to_private_key = os.path.join(os.path.expanduser("~"),".ssh/id_rsa")
    private_key = paramiko.RSAKey.from_private_key_file(os.path.expanduser(path_to_private_key))
    username = get_artemis_config_value(section=ip_address, option="username", default_generator=lambda: getpass.getuser())
    ssh_conn = paramiko.SSHClient()
    ssh_conn.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh_conn.connect(hostname=ip_address, username=username, pkey=private_key)
    return ssh_conn


def check_ssh_connection(ip_address):
    '''
    tries to load necessary information from the remote ~/.artemisrc file and execute a test remote function call. This is to verify that ssh connection is available
    :param ip_address: the ip_address to call against
    :return:
    '''

    test_function = 'python -c "import socket; print([l for l in ([ip for ip in socket.gethostbyname_ex(socket.gethostname())[2] if not ip.startswith(\'127.\')][:1], [[(s.connect((\'8.8.8.8\', 53)), s.getsockname()[0], s.close()) for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]]) if l][0][0])"'
    ssh_conn = get_ssh_connection(ip_address=ip_address)
    stdin , stdout, stderr = ssh_conn.exec_command(test_function)
    out = stdout.read().strip()
    assert out == ip_address, "The remote server resolved a different ip-address than the one this computer used to contact it. This must not be a problem, but may be worth investigating"
    err = stderr.read()
    assert not err, "The remote server could not execute the test function. It returned the following error: \n %s"%err
    ssh_conn.close()

def check_pid(pid):
    """ Check For the existence of a unix pid.
    Source: https://stackoverflow.com/a/568285/2068168"""
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True



__author__ = 'peter'


class KwargDealer(object):
    """
    Python's kwargs mechanism provides a nice way to optionally override default values.  But there are occasionally
    legitimate cases where the {existance of/appropriate defaults for} some kwargs depend on the values of others.  For
    these cases, we use KwargDealer to distribute the appropriate defaults.
    """
    def __init__(self, kwargs):
        self._kwargs = kwargs

    def deal(self, default_dict):
        parameter_dict = default_dict.copy()
        for k in default_dict:
            if k in self._kwargs:
                parameter_dict[k] = self._kwargs[k]
                del self._kwargs[k]
        return parameter_dict

    def assert_empty(self):
        assert len(self._kwargs)==0, 'Unused kwargs remain: %s' % (self._kwargs, )




import sys
__author__ = 'peter'


class TextWrappingPrinter(object):

    def __init__(self, newline_every = 70):

        self.newline_every = newline_every
        self._space_to_newline = newline_every

    def write(self, text):

        while True:
            try:
                first_newline_index = text.index('\n')
                if first_newline_index < self._space_to_newline:
                    sys.stdout.write(text[:first_newline_index+1])
                    text = text[first_newline_index+1:]
                    self._space_to_newline = self.newline_every
                else:
                    # Need to insert newline
                    sys.stdout.write(text[:self._space_to_newline] + '\n')
                    text = text[self._space_to_newline:]
                    self._space_to_newline = self.newline_every
            except ValueError:  # No newlines
                if len(text) < self._space_to_newline:  # Can put all text into line
                    sys.stdout.write(text)
                    self._space_to_newline -= len(text)
                    break
                else:
                    # Need to insert newline
                    sys.stdout.write(text[:self._space_to_newline] + '\n')
                    text = text[self._space_to_newline:]
                    self._space_to_newline = self.newline_every
